"""cdefs for message.py"""

import cython

from ._private.marshaller cimport Marshaller
from .signature cimport Variant


cdef object ErrorType
cdef object SignatureTree
cdef object SignatureType
cdef object MessageType


cdef object HEADER_PATH
cdef object HEADER_INTERFACE
cdef object HEADER_MEMBER
cdef object HEADER_ERROR_NAME
cdef object HEADER_REPLY_SERIAL
cdef object HEADER_DESTINATION
cdef object HEADER_SENDER
cdef object HEADER_SIGNATURE
cdef object HEADER_UNIX_FDS


cdef object LITTLE_ENDIAN
cdef object PROTOCOL_VERSION

cdef object MESSAGE_FLAG
cdef object MESSAGE_FLAG_NONE
cdef object MESSAGE_TYPE_METHOD_CALL

cdef get_signature_tree

cdef class Message:

    cdef public object destination
    cdef public object path
    cdef public object interface
    cdef public object member
    cdef public object message_type
    cdef public object flags
    cdef public object error_name
    cdef public object reply_serial
    cdef public object sender
    cdef public cython.list unix_fds
    cdef public object signature
    cdef public object signature_tree
    cdef public object body
    cdef public object serial

    @cython.locals(
        body_buffer=cython.bytearray,
        header_buffer=cython.bytearray
    )
    cpdef _marshall(self, object negotiate_unix_fd)



"""This tests the ability to send and receive file descriptors in dbus messages"""
import os

import pytest

from dbus_fast import Message, MessageType
from dbus_fast.aio import MessageBus
from dbus_fast.service import ServiceInterface, dbus_property, method, signal
from dbus_fast.signature import SignatureTree, Variant


def open_file():
    return os.open(os.devnull, os.O_RDONLY)


class ExampleInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)
        self.fds = []

    @method()
    def ReturnsFd(self) -> "h":
        fd = open_file()
        self.fds.append(fd)
        return fd

    @method()
    def AcceptsFd(self, fd: "h"):
        assert fd != 0
        self.fds.append(fd)

    def get_last_fd(self):
        return self.fds[-1]

    def cleanup(self):
        for fd in self.fds:
            os.close(fd)
        self.fds.clear()

    @signal()
    def SignalFd(self) -> "h":
        fd = open_file()
        self.fds.append(fd)
        return fd

    @dbus_property()
    def PropFd(self) -> "h":
        if not self.fds:
            fd = open_file()
            self.fds.append(fd)
        return self.fds[-1]

    @PropFd.setter
    def PropFd(self, fd: "h"):
        assert fd
        self.fds.append(fd)


def assert_fds_equal(fd1, fd2):
    assert fd1
    assert fd2

    stat1 = os.fstat(fd1)
    stat2 = os.fstat(fd2)

    assert stat1.st_dev == stat2.st_dev
    assert stat1.st_ino == stat2.st_ino
    assert stat1.st_rdev == stat2.st_rdev


@pytest.mark.asyncio
async def test_sending_file_descriptor_low_level():
    bus1 = await MessageBus(negotiate_unix_fd=True).connect()
    bus2 = await MessageBus(negotiate_unix_fd=True).connect()

    fd_before = open_file()
    fd_after = None

    msg = Message(
        destination=bus1.unique_name,
        path="/org/test/path",
        interface="org.test.iface",
        member="SomeMember",
        body=[0],
        signature="h",
        unix_fds=[fd_before],
    )

    def message_handler(sent):
        nonlocal fd_after
        if sent.sender == bus2.unique_name and sent.serial == msg.serial:
            assert sent.path == msg.path
            assert sent.serial == msg.serial
            assert sent.interface == msg.interface
            assert sent.member == msg.member
            assert sent.body == [0]
            assert len(sent.unix_fds) == 1
            fd_after = sent.unix_fds[0]
            bus1.send(Message.new_method_return(sent, "s", ["got it"]))
            bus1.remove_message_handler(message_handler)
            return True

    bus1.add_message_handler(message_handler)

    reply = await bus2.call(msg)
    assert reply.body == ["got it"]
    assert fd_after is not None

    assert_fds_equal(fd_before, fd_after)

    for fd in [fd_before, fd_after]:
        os.close(fd)
    for bus in [bus1, bus2]:
        bus.disconnect()


@pytest.mark.asyncio
async def test_high_level_service_fd_passing(event_loop):
    bus1 = await MessageBus(negotiate_unix_fd=True).connect()
    bus2 = await MessageBus(negotiate_unix_fd=True).connect()

    interface_name = "test.interface"
    interface = ExampleInterface(interface_name)
    export_path = "/test/path"

    async def call(member, signature="", body=[], unix_fds=[], iface=interface.name):
        return await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=export_path,
                interface=iface,
                member=member,
                signature=signature,
                body=body,
                unix_fds=unix_fds,
            )
        )

    bus1.export(export_path, interface)

    # test that an fd can be returned by the service
    reply = await call("ReturnsFd")
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body
    assert reply.signature == "h"
    assert len(reply.unix_fds) == 1
    assert_fds_equal(interface.get_last_fd(), reply.unix_fds[0])
    interface.cleanup()
    os.close(reply.unix_fds[0])

    # test that an fd can be sent to the service
    fd = open_file()
    reply = await call("AcceptsFd", signature="h", body=[0], unix_fds=[fd])
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body
    assert_fds_equal(interface.get_last_fd(), fd)

    interface.cleanup()
    os.close(fd)

    # signals
    fut = event_loop.create_future()

    def fd_listener(msg):
        if msg.sender == bus1.unique_name and msg.message_type == MessageType.SIGNAL:
            fut.set_result(msg)

    reply = await bus2.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            member="AddMatch",
            signature="s",
            body=[f"sender='{bus1.unique_name}'"],
        )
    )
    assert reply.message_type == MessageType.METHOD_RETURN

    bus2.add_message_handler(fd_listener)
    interface.SignalFd()
    reply = await fut

    assert len(reply.unix_fds) == 1
    assert reply.body == [0]
    assert_fds_equal(reply.unix_fds[0], interface.get_last_fd())

    interface.cleanup()
    os.close(reply.unix_fds[0])

    # properties
    reply = await call(
        "Get", "ss", [interface_name, "PropFd"], iface="org.freedesktop.DBus.Properties"
    )
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body
    assert reply.body[0].signature == "h"
    assert reply.body[0].value == 0
    assert len(reply.unix_fds) == 1
    assert_fds_equal(interface.get_last_fd(), reply.unix_fds[0])
    interface.cleanup()
    os.close(reply.unix_fds[0])

    fd = open_file()
    reply = await call(
        "Set",
        "ssv",
        [interface_name, "PropFd", Variant("h", 0)],
        iface="org.freedesktop.DBus.Properties",
        unix_fds=[fd],
    )
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body
    assert_fds_equal(interface.get_last_fd(), fd)
    interface.cleanup()
    os.close(fd)

    reply = await call(
        "GetAll", "s", [interface_name], iface="org.freedesktop.DBus.Properties"
    )
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body
    assert reply.body[0]["PropFd"].signature == "h"
    assert reply.body[0]["PropFd"].value == 0
    assert len(reply.unix_fds) == 1
    assert_fds_equal(interface.get_last_fd(), reply.unix_fds[0])
    interface.cleanup()
    os.close(reply.unix_fds[0])

    for bus in [bus1, bus2]:
        bus.disconnect()


@pytest.mark.asyncio
async def test_sending_file_descriptor_with_proxy(event_loop):
    name = "dbus.next.test.service"
    path = "/test/path"
    interface_name = "test.interface"

    bus = await MessageBus(negotiate_unix_fd=True).connect()

    interface = ExampleInterface(interface_name)
    bus.export(path, interface)
    await bus.request_name(name)

    intr = await bus.introspect(name, path)

    proxy = bus.get_proxy_object(name, path, intr)
    proxy_interface = proxy.get_interface(interface_name)

    # test fds are replaced correctly in all high level interfaces
    fd = await proxy_interface.call_returns_fd()
    assert_fds_equal(interface.get_last_fd(), fd)
    interface.cleanup()
    os.close(fd)

    fd = open_file()
    await proxy_interface.call_accepts_fd(fd)
    assert_fds_equal(interface.get_last_fd(), fd)
    interface.cleanup()
    os.close(fd)

    fd = await proxy_interface.get_prop_fd()
    assert_fds_equal(interface.get_last_fd(), fd)
    interface.cleanup()
    os.close(fd)

    fd = open_file()
    await proxy_interface.set_prop_fd(fd)
    assert_fds_equal(interface.get_last_fd(), fd)
    interface.cleanup()
    os.close(fd)

    fut = event_loop.create_future()

    def on_signal_fd(fd):
        fut.set_result(fd)
        proxy_interface.off_signal_fd(on_signal_fd)

    proxy_interface.on_signal_fd(on_signal_fd)
    interface.SignalFd()
    fd = await fut
    assert_fds_equal(interface.get_last_fd(), fd)
    interface.cleanup()
    os.close(fd)

    bus.disconnect()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "result, out_signature, expected",
    [
        pytest.param(5, "h", ([0], [5]), id='Signature: "h"'),
        pytest.param([5, "foo"], "hs", ([0, "foo"], [5]), id='Signature: "hs"'),
        pytest.param([5, 7], "hh", ([0, 1], [5, 7]), id='Signature: "hh"'),
        pytest.param([5, 7], "ah", ([[0, 1]], [5, 7]), id='Signature: "ah"'),
        pytest.param([9], "ah", ([[0]], [9]), id='Signature: "ah"'),
        pytest.param([3], "(h)", ([[0]], [3]), id='Signature: "(h)"'),
        pytest.param([3, "foo"], "(hs)", ([[0, "foo"]], [3]), id='Signature: "(hs)"'),
        pytest.param(
            [[7, "foo"], [8, "bar"]],
            "a(hs)",
            ([[[0, "foo"], [1, "bar"]]], [7, 8]),
            id='Signature: "a(hs)"',
        ),
        pytest.param({"foo": 3}, "a{sh}", ([{"foo": 0}], [3]), id='Signature: "a{sh}"'),
        pytest.param(
            {"foo": 3, "bar": 6},
            "a{sh}",
            ([{"foo": 0, "bar": 1}], [3, 6]),
            id='Signature: "a{sh}"',
        ),
        pytest.param(
            {"foo": [3, 8]},
            "a{sah}",
            ([{"foo": [0, 1]}], [3, 8]),
            id='Signature: "a{sah}"',
        ),
        pytest.param(
            {"foo": Variant("t", 100)},
            "a{sv}",
            ([{"foo": Variant("t", 100)}], []),
            id='Signature: "a{sv}"',
        ),
        pytest.param(
            ["one", ["two", [Variant("s", "three")]]],
            "(s(s(v)))",
            ([["one", ["two", [Variant("s", "three")]]]], []),
            id='Signature: "(s(s(v)))"',
        ),
        pytest.param(
            Variant("h", 2), "v", ([Variant("h", 0)], [2]), id='Variant with: "h"'
        ),
        pytest.param(
            Variant("(hh)", [2, 8]),
            "v",
            ([Variant("(hh)", [0, 1])], [2, 8]),
            id='Variant with: "(hh)"',
        ),
        pytest.param(
            Variant("ah", [2, 4]),
            "v",
            ([Variant("ah", [0, 1])], [2, 4]),
            id='Variant with: "ah"',
        ),
        pytest.param(
            Variant("(ss)", ["hello", "world"]),
            "v",
            ([Variant("(ss)", ["hello", "world"])], []),
            id='Variant with: "(ss)"',
        ),
        pytest.param(
            Variant("v", Variant("t", 100)),
            "v",
            ([Variant("v", Variant("t", 100))], []),
            id='Variant with: "v"',
        ),
        pytest.param(
            [
                Variant("v", Variant("(ss)", ["hello", "world"])),
                {"foo": Variant("t", 100)},
                ["one", ["two", [Variant("s", "three")]]],
            ],
            "va{sv}(s(s(v)))",
            (
                [
                    Variant("v", Variant("(ss)", ["hello", "world"])),
                    {"foo": Variant("t", 100)},
                    ["one", ["two", [Variant("s", "three")]]],
                ],
                [],
            ),
            id='Variant with: "va{sv}(s(s(v)))"',
        ),
    ],
)
async def test_fn_result_to_body(result, out_signature, expected):
    out_signature_tree = SignatureTree(out_signature)
    assert ServiceInterface._fn_result_to_body(result, out_signature_tree) == expected



import io
import json
import os
from enum import Enum
from typing import Any, Dict

import pytest

from dbus_fast import Message, MessageFlag, MessageType, SignatureTree, Variant
from dbus_fast._private._cython_compat import FakeCython
from dbus_fast._private.unmarshaller import Unmarshaller
from dbus_fast.unpack import unpack_variants


def print_buf(buf):
    i = 0
    while True:
        p = buf[i : i + 8]
        if not p:
            break
        print(p)
        i += 8


# these messages have been verified with another library
with open(os.path.dirname(__file__) + "/data/messages.json") as f:
    table = json.load(f)

with open(os.path.dirname(__file__) + "/data/get_managed_objects.hex") as fp:
    get_managed_objects_msg = fp.read()


def json_to_message(message: Dict[str, Any]) -> Message:
    copy = dict(message)
    if "message_type" in copy:
        copy["message_type"] = MessageType(copy["message_type"])
    if "flags" in copy:
        copy["flags"] = MessageFlag(copy["flags"])

    return Message(**copy)


# variants are an object in the json
def replace_variants(type_, item):
    if type_.token == "v" and type(item) is not Variant:
        item = Variant(
            item["signature"],
            replace_variants(SignatureTree(item["signature"]).types[0], item["value"]),
        )
    elif type_.token == "a":
        for i, item_child in enumerate(item):
            if type_.children[0].token == "{":
                for k, v in item.items():
                    item[k] = replace_variants(type_.children[0].children[1], v)
            else:
                item[i] = replace_variants(type_.children[0], item_child)
    elif type_.token == "(":
        for i, item_child in enumerate(item):
            if type_.children[0].token == "{":
                assert False
            else:
                item[i] = replace_variants(type_.children[i], item_child)

    return item


def json_dump(what):
    def dumper(obj):
        try:
            return obj.toJSON()
        except Exception:
            return obj.__dict__

    return json.dumps(what, default=dumper, indent=2)


def test_marshalling_with_table():
    for item in table:
        message = json_to_message(item["message"])

        body = []
        for i, type_ in enumerate(message.signature_tree.types):
            body.append(replace_variants(type_, message.body[i]))
        message.body = body

        buf = message._marshall(False)
        data = bytes.fromhex(item["data"])

        if buf != data:
            print("message:")
            print(json_dump(item["message"]))
            print("")
            print("mine:")
            print_buf(bytes(buf))
            print("")
            print("theirs:")
            print_buf(data)

        assert buf == data


@pytest.mark.parametrize("unmarshall_table", (table,))
def test_unmarshalling_with_table(unmarshall_table):
    from dbus_fast._private import unmarshaller

    for item in unmarshall_table:
        stream = io.BytesIO(bytes.fromhex(item["data"]))
        unmarshaller = Unmarshaller(stream)
        try:
            unmarshaller.unmarshall()
        except Exception as e:
            print("message failed to unmarshall:")
            print(json_dump(item["message"]))
            raise e

        message = json_to_message(item["message"])

        body = []
        for i, type_ in enumerate(message.signature_tree.types):
            body.append(replace_variants(type_, message.body[i]))
        message.body = body

        for attr in [
            "body",
            "signature",
            "message_type",
            "destination",
            "path",
            "interface",
            "member",
            "flags",
            "serial",
        ]:
            assert getattr(unmarshaller.message, attr) == getattr(
                message, attr
            ), f"attr doesnt match: {attr}"


def test_unmarshall_can_resume():
    """Verify resume works."""
    bluez_rssi_message = (
        "6c04010134000000e25389019500000001016f00250000002f6f72672f626c75657a2f686369302f6465"
        "765f30385f33415f46325f31455f32425f3631000000020173001f0000006f72672e667265656465736b"
        "746f702e444275732e50726f7065727469657300030173001100000050726f706572746965734368616e"
        "67656400000000000000080167000873617b73767d617300000007017300040000003a312e3400000000"
        "110000006f72672e626c75657a2e446576696365310000000e0000000000000004000000525353490001"
        "6e00a7ff000000000000"
    )
    message_bytes = bytes.fromhex(bluez_rssi_message)

    class SlowStream(io.IOBase):
        """A fake stream that will only give us one byte at a time."""

        def __init__(self):
            self.data = message_bytes
            self.pos = 0

        def read(self, n) -> bytes:
            data = self.data[self.pos : self.pos + 1]
            self.pos += 1
            return data

    stream = SlowStream()
    unmarshaller = Unmarshaller(stream)

    for _ in range(len(bluez_rssi_message)):
        if unmarshaller.unmarshall():
            break
    assert unmarshaller.message is not None


def test_unmarshall_bluez_message():
    bluez_mfr_message = (
        "6c040101780000009aca0a009500000001016f00250000002f6f72672f626c75657a2f686369302f646576"
        "5f44305f43325f34455f30385f41425f3537000000020173001f0000006f72672e667265656465736b746f"
        "702e444275732e50726f7065727469657300030173001100000050726f706572746965734368616e676564"
        "00000000000000080167000873617b73767d617300000007017300040000003a312e340000000011000000"
        "6f72672e626c75657a2e446576696365310000005400000000000000040000005253534900016e00aaff00"
        "00100000004d616e756661637475726572446174610005617b71767d002400000075000261790000001800"
        "00004204010170d0c24e08ab57d2c24e08ab5601000000000000000000006c040101340000009bca0a0095"
        "00000001016f002500"
    )
    message_bytes = bytes.fromhex(bluez_mfr_message)
    stream = io.BytesIO(message_bytes)
    unmarshaller = Unmarshaller(stream)
    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    assert message.body == [
        "org.bluez.Device1",
        {
            "ManufacturerData": Variant(
                "a{qv}",
                {
                    117: Variant(
                        "ay",
                        bytearray(
                            b"B\x04\x01\x01p\xd0\xc2N\x08\xabW\xd2\xc2N\x08\xabV\x01\x00\x00\x00\x00\x00\x00"
                        ),
                    )
                },
            ),
            "RSSI": Variant("n", -86),
        },
        [],
    ]
    assert message.sender == ":1.4"
    assert message.path == "/org/bluez/hci0/dev_D0_C2_4E_08_AB_57"
    assert message.interface == "org.freedesktop.DBus.Properties"
    assert message.member == "PropertiesChanged"
    assert message.signature == "sa{sv}as"
    assert message.message_type == MessageType.SIGNAL
    assert message.flags == MessageFlag.NO_REPLY_EXPECTED
    assert message.serial == 707226
    assert message.destination is None
    unpacked = unpack_variants(message.body)
    assert unpacked == [
        "org.bluez.Device1",
        {
            "ManufacturerData": {
                117: bytearray(
                    b"B\x04\x01\x01p\xd0\xc2N\x08\xabW\xd2"
                    b"\xc2N\x08\xabV\x01\x00\x00"
                    b"\x00\x00\x00\x00"
                )
            },
            "RSSI": -86,
        },
        [],
    ]


def test_unmarshall_bluez_interfaces_added_message():
    bluez_interfaces_added_message = (
        b'l\4\1\1\240\2\0\0\227\272\23\0u\0\0\0\1\1o\0\1\0\0\0/\0\0\0\0\0\0\0\2\1s\0"\0\0\0'
        b"org.freedesktop.DBus.ObjectManager\0\0\0\0\0\0\3\1s\0\17\0\0\0InterfacesAdded\0\10"
        b"\1g\0\noa{sa{sv}}\0\7\1s\0\4\0\0\0:1.4\0\0\0\0%\0\0\0/org/bluez/hci1/dev_58_2D_34"
        b"_60_26_36\0\0\0p\2\0\0#\0\0\0org.freedesktop.DBus.Introspectable\0\0\0\0\0\0\0\0\0"
        b"\21\0\0\0org.bluez.Device1\0\0\0\364\1\0\0\0\0\0\0\7\0\0\0Address\0\1s\0\0\21\0\0"
        b"\00058:2D:34:60:26:36\0\0\0\v\0\0\0AddressType\0\1s\0\0\6\0\0\0public\0\0\4\0\0\0"
        b"Name\0\1s\0\33\0\0\0Qingping Door/Window Sensor\0\0\0\0\0\5\0\0\0Alias\0\1s\0\0\0"
        b"\0\33\0\0\0Qingping Door/Window Sensor\0\6\0\0\0Paired\0\1b\0\0\0\0\0\0\0\0\0\0\0"
        b"\7\0\0\0Trusted\0\1b\0\0\0\0\0\0\0\0\0\0\7\0\0\0Blocked\0\1b\0\0\0\0\0\0\0\0\0\0\r"
        b"\0\0\0LegacyPairing\0\1b\0\0\0\0\0\0\0\0\0\0\0\0\4\0\0\0RSSI\0\1n\0\316\377\0\0\t"
        b"\0\0\0Connected\0\1b\0\0\0\0\0\0\0\0\5\0\0\0UUIDs\0\2as\0\0\0\0\0\0\0\0\0\0\0\7\0"
        b"\0\0Adapter\0\1o\0\0\17\0\0\0/org/bluez/hci1\0\0\0\0\0\v\0\0\0ServiceData\0\5a{sv}"
        b"\0\0@\0\0\0\0\0\0\0$\0\0\0000000fe95-0000-1000-8000-00805f9b34fb\0\2ay\0\0\0\0\f\0"
        b"\0\0000X\326\3\0026&`4-X\10\20\0\0\0ServicesResolved\0\1b\0\0\0\0\0\0\0\0\0\37\0\0"
        b"\0org.freedesktop.DBus.Properties\0\0\0\0\0"
    )

    stream = io.BytesIO(bluez_interfaces_added_message)
    unmarshaller = Unmarshaller(stream)
    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    assert message.body == [
        "/org/bluez/hci1/dev_58_2D_34_60_26_36",
        {
            "org.bluez.Device1": {
                "Adapter": Variant("o", "/org/bluez/hci1"),
                "Address": Variant("s", "58:2D:34:60:26:36"),
                "AddressType": Variant("s", "public"),
                "Alias": Variant("s", "Qingping Door/Window Sensor"),
                "Blocked": Variant("b", False),
                "Connected": Variant("b", False),
                "LegacyPairing": Variant("b", False),
                "Name": Variant("s", "Qingping Door/Window Sensor"),
                "Paired": Variant("b", False),
                "RSSI": Variant("n", -50),
                "ServiceData": Variant(
                    "a{sv}",
                    {
                        "0000fe95-0000-1000-8000-00805f9b34fb": Variant(
                            "ay", bytearray(b"0X\xd6\x03\x026&`4-X\x08")
                        )
                    },
                ),
                "ServicesResolved": Variant("b", False),
                "Trusted": Variant("b", False),
                "UUIDs": Variant("as", []),
            },
            "org.freedesktop.DBus.Introspectable": {},
            "org.freedesktop.DBus.Properties": {},
        },
    ]
    assert message.sender == ":1.4"
    assert message.path == "/"
    assert message.interface == "org.freedesktop.DBus.ObjectManager"
    assert message.member == "InterfacesAdded"
    assert message.signature == "oa{sa{sv}}"
    assert message.message_type == MessageType.SIGNAL
    assert message.flags == MessageFlag.NO_REPLY_EXPECTED
    assert message.serial == 1292951
    assert message.destination is None
    unpacked = unpack_variants(message.body)
    assert unpacked == [
        "/org/bluez/hci1/dev_58_2D_34_60_26_36",
        {
            "org.bluez.Device1": {
                "Adapter": "/org/bluez/hci1",
                "Address": "58:2D:34:60:26:36",
                "AddressType": "public",
                "Alias": "Qingping Door/Window Sensor",
                "Blocked": False,
                "Connected": False,
                "LegacyPairing": False,
                "Name": "Qingping Door/Window Sensor",
                "Paired": False,
                "RSSI": -50,
                "ServiceData": {
                    "0000fe95-0000-1000-8000-00805f9b34fb": bytearray(
                        b"0X\xd6\x03" b"\x026&`" b"4-X\x08"
                    )
                },
                "ServicesResolved": False,
                "Trusted": False,
                "UUIDs": [],
            },
            "org.freedesktop.DBus.Introspectable": {},
            "org.freedesktop.DBus.Properties": {},
        },
    ]


def test_unmarshall_bluez_interfaces_removed_message():
    bluez_interfaces_removed_message = (
        b'l\4\1\1\222\0\0\0\377@-\0~\0\0\0\1\1o\0\1\0\0\0/\0\0\0\0\0\0\0\2\1s\0"\0\0\0'
        b"org.freedesktop.DBus.ObjectManager\0\0\0\0\0\0\3\1s\0\21\0\0\0InterfacesRemoved"
        b"\0\0\0\0\0\0\0\10\1g\0\3oas\0\0\0\0\0\0\0\0\7\1s\0\5\0\0\0:1.12\0\0\0%\0\0\0"
        b"/org/bluez/hci0/dev_5F_13_47_38_26_55\0\0\0b\0\0\0\37\0\0\0org.freedesktop.DBus"
        b".Properties\0#\0\0\0org.freedesktop.DBus.Introspectable\0\21\0\0\0org.bluez.Dev"
        b"ice1\0"
    )

    stream = io.BytesIO(bluez_interfaces_removed_message)
    unmarshaller = Unmarshaller(stream)
    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    assert message.body == [
        "/org/bluez/hci0/dev_5F_13_47_38_26_55",
        [
            "org.freedesktop.DBus.Properties",
            "org.freedesktop.DBus.Introspectable",
            "org.bluez.Device1",
        ],
    ]
    assert message.sender == ":1.12"
    assert message.path == "/"
    assert message.interface == "org.freedesktop.DBus.ObjectManager"
    assert message.member == "InterfacesRemoved"
    assert message.signature == "oas"
    assert message.message_type == MessageType.SIGNAL
    assert message.flags == MessageFlag.NO_REPLY_EXPECTED
    assert message.serial == 2965759
    assert message.destination is None
    unpacked = unpack_variants(message.body)
    assert unpacked == [
        "/org/bluez/hci0/dev_5F_13_47_38_26_55",
        [
            "org.freedesktop.DBus.Properties",
            "org.freedesktop.DBus.Introspectable",
            "org.bluez.Device1",
        ],
    ]


def test_unmarshall_bluez_properties_changed_with_service_data():
    bluez_properties_changed_message = (
        b"l\4\1\1\334\0\0\0@\236.\0\226\0\0\0\1\1o\0%\0\0\0/org/bluez/hci0/dev_58_2D_34_60_DA_1F"
        b"\0\0\0\2\1s\0\37\0\0\0org.freedesktop.DBus.Properties\0\3\1s\0\21\0\0\0PropertiesChanged"
        b"\0\0\0\0\0\0\0\10\1g\0\10sa{sv}as\0\0\0\7\1s\0\5\0\0\0:1.12\0\0\0\21\0\0\0org.bluez.Devi"
        b"ce1\0\0\0\270\0\0\0\0\0\0\0\4\0\0\0RSSI\0\1n\0\301\377\0\0\v\0\0\0ServiceData\0\5a{sv}"
        b"\0\0\210\0\0\0\0\0\0\0$\0\0\0000000fdcd-0000-1000-8000-00805f9b34fb\0\2ay\0\0\0\0\24\0"
        b"\0\0\10\22\37\332`4-X\2\1U\17\1\315\t\4\5\0\0\0$\0\0\0000000fe95-0000-1000-8000-00805f"
        b"9b34fb\0\2ay\0\0\0\0\f\0\0\0000X\203\n\2\37\332`4-X\10\0\0\0\0"
    )

    stream = io.BytesIO(bluez_properties_changed_message)
    unmarshaller = Unmarshaller(stream)
    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    assert message.body == [
        "org.bluez.Device1",
        {
            "RSSI": Variant("n", -63),
            "ServiceData": Variant(
                "a{sv}",
                {
                    "0000fdcd-0000-1000-8000-00805f9b34fb": Variant(
                        "ay",
                        bytearray(
                            b"\x08\x12\x1f\xda`4-X\x02\x01U\x0f\x01\xcd\t\x04\x05\x00\x00\x00"
                        ),
                    ),
                    "0000fe95-0000-1000-8000-00805f9b34fb": Variant(
                        "ay", bytearray(b"0X\x83\n\x02\x1f\xda`4-X\x08")
                    ),
                },
            ),
        },
        [],
    ]
    assert message.sender == ":1.12"
    assert message.path == "/org/bluez/hci0/dev_58_2D_34_60_DA_1F"
    assert message.interface == "org.freedesktop.DBus.Properties"
    assert message.member == "PropertiesChanged"
    assert message.signature == "sa{sv}as"
    assert message.message_type == MessageType.SIGNAL
    assert message.flags == MessageFlag.NO_REPLY_EXPECTED
    assert message.serial == 3055168
    assert message.destination is None
    unpacked = unpack_variants(message.body)
    assert unpacked == [
        "org.bluez.Device1",
        {
            "RSSI": -63,
            "ServiceData": {
                "0000fdcd-0000-1000-8000-00805f9b34fb": bytearray(
                    b"\x08\x12\x1f\xda"
                    b"`4-X"
                    b"\x02\x01U\x0f"
                    b"\x01\xcd\t\x04"
                    b"\x05\x00\x00\x00"
                ),
                "0000fe95-0000-1000-8000-00805f9b34fb": bytearray(
                    b"0X\x83\n" b"\x02\x1f\xda`" b"4-X\x08"
                ),
            },
        },
        [],
    ]


def test_unmarshall_multiple_messages():
    """Test we can unmarshall multiple messages in a single packet."""
    multiple_message_packet = (
        b"l\4\1\0014\0\0\0J\27\230\0\225\0\0\0\1\1o\0%\0\0\0/org/bluez/hci0/dev_F0_B3_EC_15_7F_8E\0\0\0\2\1s\0\37\0\0\0"
        b"org.freedesktop.DBus.Properties\0\3\1s\0\21\0\0\0PropertiesChanged\0\0\0\0\0\0\0\10\1g\0\10sa{sv}as\0\0\0\7\1"
        b"s\0\4\0\0\0:1.4\0\0\0\0\21\0\0\0org.bluez.Device1\0\0\0\16\0\0\0\0\0\0\0\4\0\0\0RSSI\0\1n\0\264\377\0\0\0\0\0"
        b"\0l\4\1\0014\0\0\0K\27\230\0\225\0\0\0\1\1o\0%\0\0\0/org/bluez/hci0/dev_3F_70_98_0F_08_CB\0\0\0\2\1s\0\37\0\0"
        b"\0org.freedesktop.DBus.Properties\0\3\1s\0\21\0\0\0PropertiesChanged\0\0\0\0\0\0\0\10\1g\0\10sa{sv}as\0\0\0\7"
        b"\1s\0\4\0\0\0:1.4\0\0\0\0\21\0\0\0org.bluez.Device1\0\0\0\16\0\0\0\0\0\0\0\4\0\0\0RSSI\0\1n\0\260\377\0\0\0\0"
        b"\0\0l\4\1\0014\0\0\0L\27\230\0\225\0\0\0\1\1o\0%\0\0\0/org/bluez/hci0/dev_D8_35_67_A4_F5_A5\0\0\0\2\1s\0\37\0"
        b"\0\0org.freedesktop.DBus.Properties\0\3\1s\0\21\0\0\0PropertiesChanged\0\0\0\0\0\0\0\10\1g\0\10sa{sv}as\0\0\0"
        b"\7\1s\0\4\0\0\0:1.4\0\0\0\0\21\0\0\0org.bluez.Device1\0\0\0\16\0\0\0\0\0\0\0\4\0\0\0RSSI\0\1n\0\242\377\0\0\0"
        b"\0\0\0"
    )
    stream = io.BytesIO(multiple_message_packet)
    unmarshaller = Unmarshaller(stream)
    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    unpacked = unpack_variants(message.body)
    assert unpacked == ["org.bluez.Device1", {"RSSI": -76}, []]

    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    unpacked = unpack_variants(message.body)
    assert unpacked == ["org.bluez.Device1", {"RSSI": -80}, []]

    assert unmarshaller.unmarshall()
    message = unmarshaller.message
    assert message is not None
    unpacked = unpack_variants(message.body)
    assert unpacked == ["org.bluez.Device1", {"RSSI": -94}, []]

    with pytest.raises(EOFError):
        unmarshaller.unmarshall()


def test_ay_buffer():
    body = [bytes(10000)]
    msg = Message(path="/test", member="test", signature="ay", body=body)
    marshalled = msg._marshall(False)
    unmarshalled_msg = Unmarshaller(io.BytesIO(marshalled)).unmarshall()
    assert unmarshalled_msg.body[0] == body[0]


def tests_fallback_no_cython():
    assert FakeCython().compiled is False


def test_unmarshall_large_message():
    stream = io.BytesIO(bytes.fromhex(get_managed_objects_msg))
    unmarshaller = Unmarshaller(stream)
    unmarshaller.unmarshall()
    message = unmarshaller.message
    unpacked = unpack_variants(message.body)
    objects = unpacked[0]
    assert objects["/org/bluez/hci0"] == {
        "org.bluez.Adapter1": {
            "Address": "00:1A:7D:DA:71:04",
            "AddressType": "public",
            "Alias": "homeassistant",
            "Class": 2883584,
            "Discoverable": False,
            "DiscoverableTimeout": 180,
            "Discovering": True,
            "Modalias": "usb:v1D6Bp0246d053F",
            "Name": "homeassistant",
            "Pairable": False,
            "PairableTimeout": 0,
            "Powered": True,
            "Roles": ["central", "peripheral"],
            "UUIDs": [
                "0000110e-0000-1000-8000-00805f9b34fb",
                "0000110a-0000-1000-8000-00805f9b34fb",
                "00001200-0000-1000-8000-00805f9b34fb",
                "0000110b-0000-1000-8000-00805f9b34fb",
                "00001108-0000-1000-8000-00805f9b34fb",
                "0000110c-0000-1000-8000-00805f9b34fb",
                "00001800-0000-1000-8000-00805f9b34fb",
                "00001801-0000-1000-8000-00805f9b34fb",
                "0000180a-0000-1000-8000-00805f9b34fb",
                "00001112-0000-1000-8000-00805f9b34fb",
            ],
        },
        "org.bluez.GattManager1": {},
        "org.bluez.LEAdvertisingManager1": {
            "ActiveInstances": 0,
            "SupportedIncludes": ["tx-power", "appearance", "local-name"],
            "SupportedInstances": 5,
        },
        "org.bluez.Media1": {},
        "org.bluez.NetworkServer1": {},
        "org.freedesktop.DBus.Introspectable": {},
        "org.freedesktop.DBus.Properties": {},
    }
    assert objects["/org/bluez/hci0/dev_CD_A3_FA_D1_50_56/service000b/char000c"] == {
        "org.bluez.GattCharacteristic1": {
            "Flags": ["read"],
            "Service": "/org/bluez/hci0/dev_CD_A3_FA_D1_50_56/service000b",
            "UUID": "e604e95d-a759-4817-87d3-aa005083a0d1",
            "Value": bytearray(b""),
        },
        "org.freedesktop.DBus.Introspectable": {},
        "org.freedesktop.DBus.Properties": {},
    }


def test_unmarshall_big_endian_message():
    """Test we can unmarshall a big endian message."""
    msg = (
        b"B\x01\x00\x01\x00\x00\x00 \x00\x00\x00\x00\x00\x00\x00\x82"
        b"\x01\x01o\x00\x00\x00\x00\x01/\x00\x00\x00\x00\x00\x00\x00"
        b'\x02\x01s\x00\x00\x00\x00"org.freedesktop.DBus.ObjectManag'
        b"er\x00\x00\x00\x00\x00\x00\x03\x01s\x00\x00\x00\x00\x11GetManagedOb"
        b"jects\x00\x00\x00\x00\x00\x00\x00\x06\x01s\x00\x00\x00\x00\torg."
        b"bluez\x00\x00\x00\x00\x00\x00\x00\x08\x01g\x00\x04ussv\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00*\x00\x00\x00\x03zip\x00"
        b"\x00\x00\x00\x07Trusted\x00\x01b\x00\x00\x00\x00\x00\x01"
    )

    stream = io.BytesIO(msg)
    unmarshaller = Unmarshaller(stream)
    unmarshaller.unmarshall()
    message = unmarshaller.message
    unpacked = unpack_variants(message.body)
    assert unpacked == [42, "zip", "Trusted", True]


class RaucState(str, Enum):
    """Rauc slot states."""

    GOOD = "good"
    BAD = "bad"
    ACTIVE = "active"


def test_marshalling_enum():
    """Test marshalling an enum."""
    msg = Message(
        path="/test",
        member="test",
        signature="s",
        body=[RaucState.GOOD],
    )
    marshalled = msg._marshall(False)
    unmarshalled_msg = Unmarshaller(io.BytesIO(marshalled)).unmarshall()
    assert unpack_variants(unmarshalled_msg.body)[0] == RaucState.GOOD.value


def test_unmarshall_bluez_passive_message():
    """Test we can unmarshall a bluez passive message."""

    bluez_passive_message = (
        b"l\1\1\1*\0\0\0\205D\267\3\215\0\0\0\1\1o\0\35\0\0\0/org/bleak/61/281472597302272\0\0\0\6\1s\0\7\0\0"
        b"\0:1.1450\0\2\1s\0\37\0\0\0org.bluez.AdvertisementMonitor1\0\3\1s\0\v\0\0\0DeviceFound\0\0\0\0\0\10"
        b"\1g\0\1o\0\0\7\1s\0\4\0\0\0:1.4\0\0\0\0%\0\0\0/org/bluez/hci0/dev_58_D3_49_E6_02_6E\0l\1\1\1*\0\0\0"
        b"\206D\267\3\215\0\0\0\1\1o\0\35\0\0\0/org/bleak/61/281472593362560\0\0\0\6\1s\0\7\0\0\0:1.1450\0\2"
        b"\1s\0\37\0\0\0org.bluez.AdvertisementMonitor1\0\3\1s\0\v\0\0\0DeviceFound\0\0\0\0\0\10\1g\0\1o\0\0"
        b"\7\1s\0\4\0\0\0:1.4\0\0\0\0%\0\0\0/org/bluez/hci1/dev_58_D3_49_E6_02_6E\0"
    )

    stream = io.BytesIO(bluez_passive_message)
    unmarshaller = Unmarshaller(stream)
    unmarshaller.unmarshall()
    message = unmarshaller.message
    assert "/org/bluez/hci0/dev_58_D3_49_E6_02_6E" in str(message)
    unpacked = unpack_variants(message.body)
    assert unpacked == ["/org/bluez/hci0/dev_58_D3_49_E6_02_6E"]


def test_unmarshall_mount_message():
    """Test we mount message unmarshall."""

    mount_message = (
        b"l\1\0\1\30\1\0\0\213\1\0\0\266\0\0\0\1\1o\0\31\0\0\0/org/freedesktop/systemd1"
        b"\0\0\0\0\0\0\0\2\1s\0 \0\0\0org.freedesktop.systemd1.Manager\0\0\0\0\0\0\0\0"
        b"\3\1s\0\22\0\0\0StartTransientUnit\0\0\0\0\0\0\6\1s\0\30\0\0\0org.freedesktop"
        b".systemd1\0\0\0\0\0\0\0\0\10\1g\0\20ssa(sv)a(sa(sv))\0\0\0)\0\0\0mnt-data-sup"
        b"ervisor-mounts-test1234.mount\0\0\0\4\0\0\0fail\0\0\0\0\314\0\0\0\7\0\0\0Opti"
        b"ons\0\1s\0\0I\0\0\0noserverino,credentials=/mnt/data/supervisor/.mounts_crede"
        b"ntials/test1234\0\0\0\4\0\0\0Type\0\1s\0\4\0\0\0cifs\0\0\0\0\v\0\0\0Descripti"
        b"on\0\1s\0\0\37\0\0\0Supervisor cifs mount: test1234\0\4\0\0\0What\0\1s\0\v\0\0"
        b"\0//\303\274ber/test\0\0\0\0\0\0\0\0\0\0\0\0"
    )

    stream = io.BytesIO(mount_message)
    unmarshaller = Unmarshaller(stream)
    unmarshaller.unmarshall()
    message = unmarshaller.message
    assert unmarshaller.message.signature == "ssa(sv)a(sa(sv))"
    unpacked = unpack_variants(message.body)
    assert unpacked == [
        "mnt-data-supervisor-mounts-test1234.mount",
        "fail",
        [
            [
                "Options",
                "noserverino,credentials=/mnt/data/supervisor/.mounts_credentials/test1234",
            ],
            ["Type", "cifs"],
            ["Description", "Supervisor cifs mount: test1234"],
            ["What", "//über/tes"],
        ],
        [],
    ]


def test_unmarshall_mount_message_2():
    """Test we mount message unmarshall version 2."""

    mount_message = (
        b"l\1\0\1 \1\0\0+\6\0\0\266\0\0\0\1\1o\0\31\0\0\0/org/freedesktop/systemd1"
        b"\0\0\0\0\0\0\0\2\1s\0 \0\0\0org.freedesktop.systemd1.Manager\0\0\0\0\0\0"
        b"\0\0\3\1s\0\22\0\0\0StartTransientUnit\0\0\0\0\0\0\6\1s\0\30\0\0\0org.fr"
        b"eedesktop.systemd1\0\0\0\0\0\0\0\0\10\1g\0\20ssa(sv)a(sa(sv))\0\0\0(\0\0"
        b"\0mnt-data-supervisor-mounts-BadTest.mount\0\0\0\0\4\0\0\0fail\0\0\0\0\324"
        b"\0\0\0\7\0\0\0Options\0\1s\0\0H\0\0\0noserverino,credentials=/mnt/data/su"
        b"pervisor/.mounts_credentials/BadTest\0\0\0\0\4\0\0\0Type\0\1s\0\4\0\0\0cifs"
        b"\0\0\0\0\v\0\0\0Description\0\1s\0\0\36\0\0\0Supervisor cifs mount: BadTest"
        b"\0\0\4\0\0\0What\0\1s\0\23\0\0\0//doesntmatter/\303\274ber\0\0\0\0\0\0\0\0\0"
        b"\0\0\0"
    )

    stream = io.BytesIO(mount_message)
    unmarshaller = Unmarshaller(stream)
    unmarshaller.unmarshall()
    message = unmarshaller.message
    assert unmarshaller.message.signature == "ssa(sv)a(sa(sv))"
    unpacked = unpack_variants(message.body)
    assert unpacked == [
        "mnt-data-supervisor-mounts-BadTest.mount",
        "fail",
        [
            [
                "Options",
                "noserverino,credentials=/mnt/data/supervisor/.mounts_credentials/BadTest",
            ],
            ["Type", "cifs"],
            ["Description", "Supervisor cifs mount: BadTest"],
            ["What", "//doesntmatter/übe"],
        ],
        [],
    ]


def test_unmarshall_multi_byte_string():
    """Test unmarshall a multi-byte string."""

    mount_message = (
        b"l\x01\x00\x01\x1d\x00\x00\x00"
        b"\x01\x00\x00\x00x\x00\x00\x00"
        b"\x01\x01o\x00\x15\x00\x00\x00"
        b"/org/fre"
        b"edesktop"
        b"/DBus\x00\x00\x00"
        b"\x02\x01s\x00\x14\x00\x00\x00"
        b"org.free"
        b"desktop."
        b"DBus\x00\x00\x00\x00"
        b"\x03\x01s\x00\x05\x00\x00\x00"
        b"Hello\x00\x00\x00"
        b"\x06\x01s\x00\x14\x00\x00\x00"
        b"org.free"
        b"desktop."
        b"DBus\x00\x00\x00\x00"
        b"\x08\x01g\x00\x02as\x00"
        b"\x19\x00\x00\x00\x14\x00\x00\x00"
        b"//doesnt"
        b"matter/\xc3"
        b"\xbcber\x00"
    )

    stream = io.BytesIO(mount_message)
    unmarshaller = Unmarshaller(stream)
    unmarshaller.unmarshall()
    message = unmarshaller.message
    assert unmarshaller.message.signature == "as"
    unpacked = unpack_variants(message.body)
    assert unpacked == [["//doesntmatter/über"]]


def test_marshalling_struct_accepts_tuples():
    """Test marshalling a struct accepts tuples."""
    msg = Message(
        path="/test",
        member="test",
        signature="(s)",
        body=[(RaucState.GOOD,)],
    )
    marshalled = msg._marshall(False)
    unmarshalled_msg = Unmarshaller(io.BytesIO(marshalled)).unmarshall()
    assert unpack_variants(unmarshalled_msg.body)[0] == [RaucState.GOOD.value]


def test_marshalling_struct_accepts_lists():
    """Test marshalling a struct accepts lists."""
    msg = Message(
        path="/test",
        member="test",
        signature="(s)",
        body=[[RaucState.GOOD]],
    )
    marshalled = msg._marshall(False)
    unmarshalled_msg = Unmarshaller(io.BytesIO(marshalled)).unmarshall()
    assert unpack_variants(unmarshalled_msg.body)[0] == [RaucState.GOOD.value]



"""cdefs for marshaller.py"""

import cython

from ..signature cimport SignatureTree, SignatureType, Variant


cdef object PACK_UINT32

cdef bytes PACKED_UINT32_ZERO
cdef bytes PACKED_BOOL_TRUE
cdef bytes PACKED_BOOL_FALSE

cdef get_signature_tree

cdef class Marshaller:

    cdef SignatureTree signature_tree
    cdef bytearray _buf
    cdef cython.list body

    cdef _buffer(self)

    cpdef align(self, unsigned int n)

    @cython.locals(
        offset=cython.ulong,
    )
    cdef unsigned int _align(self, unsigned int n)

    cpdef write_boolean(self, object boolean, SignatureType type_)

    @cython.locals(
        written=cython.uint,
    )
    cdef unsigned int _write_boolean(self, object boolean)

    cpdef write_string(self, object value, SignatureType type_)

    @cython.locals(
        buf=cython.bytearray,
        value_len=cython.uint,
        signature_len=cython.uint,
        written=cython.uint,
    )
    cdef unsigned int _write_string(self, object value)

    @cython.locals(
        signature_len=cython.uint,
    )
    cdef unsigned int _write_signature(self, bytes signature_bytes)

    cpdef write_array(self, object array, SignatureType type_)

    @cython.locals(
        array_len=cython.uint,
        buf=cython.bytearray,
        written=cython.uint,
        token=cython.str,
        child_type=SignatureType,
        array_len_packed=cython.bytes,
        size=cython.uint,
        writer=cython.object,
        packer=cython.object,
        i=cython.uint,
    )
    cdef unsigned int _write_array(self, object array, SignatureType type_)

    cpdef write_struct(self, object array, SignatureType type_)

    @cython.locals(
        written=cython.uint,
        i=cython.uint,
    )
    cdef unsigned int _write_struct(self, object array, SignatureType type_)

    cpdef write_variant(self, Variant variant, SignatureType type_)

    @cython.locals(
        written=cython.uint,
        signature=cython.str,
        signature_bytes=cython.bytes,
    )
    cdef unsigned int _write_variant(self, Variant variant, SignatureType type_)

    @cython.locals(
        written=cython.uint,
        size=cython.uint,
    )
    cdef unsigned int _write_single(self, SignatureType type_, object body)

    @cython.locals(
        written=cython.uint,
        t=cython.str,
    )
    cpdef write_dict_entry(self, cython.list dict_entry, SignatureType type_)

    cpdef marshall(self)

    cdef _marshall(self)

    @cython.locals(
        offset=cython.ulong,
        t=cython.str,
        size=cython.uint,
        writer=cython.object,
        packer=cython.object,
        type_=SignatureType,
    )
    cdef _construct_buffer(self)



"""Stub for when Cython is not available."""


class FakeCython:
    """Stub for when Cython is not available."""

    @property
    def compiled(self) -> bool:
        return False


FAKE_CYTHON = FakeCython()



import pytest

from dbus_fast import Message
from dbus_fast.aio import MessageBus
from dbus_fast.constants import RequestNameReply
from dbus_fast.introspection import Node
from dbus_fast.service import ServiceInterface, signal
from dbus_fast.signature import Variant


class ExampleInterface(ServiceInterface):
    def __init__(self):
        super().__init__("test.interface")

    @signal()
    def SomeSignal(self) -> "s":
        return "hello"

    @signal()
    def SignalMultiple(self) -> "ss":
        return ["hello", "world"]

    @signal()
    def SignalComplex(self) -> "a{sv}":
        """Broadcast a complex signal."""
        return {"hello": Variant("s", "world")}


@pytest.mark.asyncio
async def test_signals():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    bus_intr = await bus1.introspect("org.freedesktop.DBus", "/org/freedesktop/DBus")
    bus_obj = bus1.get_proxy_object(
        "org.freedesktop.DBus", "/org/freedesktop/DBus", bus_intr
    )
    stats = bus_obj.get_interface("org.freedesktop.DBus.Debug.Stats")

    await bus1.request_name("test.signals.name")
    service_interface = ExampleInterface()
    bus1.export("/test/path", service_interface)

    obj = bus2.get_proxy_object(
        "test.signals.name", "/test/path", bus1._introspect_export_path("/test/path")
    )
    interface = obj.get_interface(service_interface.name)

    async def ping():
        await bus2.call(
            Message(
                destination=bus1.unique_name,
                interface="org.freedesktop.DBus.Peer",
                path="/test/path",
                member="Ping",
            )
        )

    err = None

    single_counter = 0

    def single_handler(value):
        try:
            nonlocal single_counter
            nonlocal err
            assert value == "hello"
            single_counter += 1
        except Exception as e:
            err = e

    multiple_counter = 0

    def multiple_handler(value1, value2):
        nonlocal multiple_counter
        nonlocal err
        try:
            assert value1 == "hello"
            assert value2 == "world"
            multiple_counter += 1
        except Exception as e:
            err = e

    await ping()
    match_rules = await stats.call_get_all_match_rules()
    assert bus2.unique_name in match_rules
    bus_match_rules = match_rules[bus2.unique_name]
    # the bus connection itself takes a rule on NameOwnerChange after the high
    # level client is initialized
    assert len(bus_match_rules) == 1
    assert len(bus2._user_message_handlers) == 0

    interface.on_some_signal(single_handler)
    interface.on_signal_multiple(multiple_handler)

    # Interlude: adding a signal handler with `on_[signal]` should add a match rule and
    # message handler. Removing a signal handler with `off_[signal]` should
    # remove the match rule and message handler to avoid memory leaks.
    await ping()
    match_rules = await stats.call_get_all_match_rules()
    assert bus2.unique_name in match_rules
    bus_match_rules = match_rules[bus2.unique_name]
    # test the match rule and user handler has been added
    assert len(bus_match_rules) == 2
    assert (
        "type='signal',interface='test.interface',path='/test/path',sender='test.signals.name'"
        in bus_match_rules
    )
    assert len(bus2._user_message_handlers) == 1

    service_interface.SomeSignal()
    await ping()
    assert err is None
    assert single_counter == 1

    service_interface.SignalMultiple()
    await ping()
    assert err is None
    assert multiple_counter == 1

    # special case: another bus with the same path and interface but on a
    # different name and connection will trigger the match rule of the first
    # (happens with mpris)
    bus3 = await MessageBus().connect()
    await bus3.request_name("test.signals.name2")
    service_interface2 = ExampleInterface()
    bus3.export("/test/path", service_interface2)

    obj = bus2.get_proxy_object(
        "test.signals.name2", "/test/path", bus3._introspect_export_path("/test/path")
    )
    # we have to add a dummy handler to add the match rule
    iface2 = obj.get_interface(service_interface2.name)

    def dummy_signal_handler(what):
        pass

    iface2.on_some_signal(dummy_signal_handler)
    await ping()

    service_interface2.SomeSignal()
    await ping()
    # single_counter is not incremented for signals of the second interface
    assert single_counter == 1

    interface.off_some_signal(single_handler)
    interface.off_signal_multiple(multiple_handler)
    iface2.off_some_signal(dummy_signal_handler)

    # After `off_[signal]`, the match rule and user handler should be removed
    await ping()
    match_rules = await stats.call_get_all_match_rules()
    assert bus2.unique_name in match_rules
    bus_match_rules = match_rules[bus2.unique_name]
    assert len(bus_match_rules) == 1
    assert (
        "type='signal',interface='test.interface',path='/test/path',sender='test.signals.name'"
        not in bus_match_rules
    )
    assert len(bus2._user_message_handlers) == 0

    bus1.disconnect()
    bus2.disconnect()
    bus3.disconnect()


@pytest.mark.asyncio
async def test_complex_signals():
    """Test complex signals with and without signature removal."""
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    await bus1.request_name("test.signals.name")
    service_interface = ExampleInterface()
    bus1.export("/test/path", service_interface)

    obj = bus2.get_proxy_object(
        "test.signals.name", "/test/path", bus1._introspect_export_path("/test/path")
    )
    interface = obj.get_interface(service_interface.name)

    async def ping():
        await bus2.call(
            Message(
                destination=bus1.unique_name,
                interface="org.freedesktop.DBus.Peer",
                path="/test/path",
                member="Ping",
            )
        )

    sig_handler_counter = 0
    sig_handler_err = None
    no_sig_handler_counter = 0
    no_sig_handler_err = None

    def complex_handler_with_sig(value):
        nonlocal sig_handler_counter
        nonlocal sig_handler_err
        try:
            assert value == {"hello": Variant("s", "world")}
            sig_handler_counter += 1
        except AssertionError as ex:
            sig_handler_err = ex

    def complex_handler_no_sig(value):
        nonlocal no_sig_handler_counter
        nonlocal no_sig_handler_err
        try:
            assert value == {"hello": "world"}
            no_sig_handler_counter += 1
        except AssertionError as ex:
            no_sig_handler_err = ex

    interface.on_signal_complex(complex_handler_with_sig)
    interface.on_signal_complex(complex_handler_no_sig, unpack_variants=True)
    await ping()

    service_interface.SignalComplex()
    await ping()
    assert sig_handler_err is None
    assert sig_handler_counter == 1
    assert no_sig_handler_err is None
    assert no_sig_handler_counter == 1

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_varargs_callback():
    """Test varargs callback for signal."""
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    await bus1.request_name("test.signals.name")
    service_interface = ExampleInterface()
    bus1.export("/test/path", service_interface)

    obj = bus2.get_proxy_object(
        "test.signals.name", "/test/path", bus1._introspect_export_path("/test/path")
    )
    interface = obj.get_interface(service_interface.name)

    async def ping():
        await bus2.call(
            Message(
                destination=bus1.unique_name,
                interface="org.freedesktop.DBus.Peer",
                path="/test/path",
                member="Ping",
            )
        )

    varargs_handler_counter = 0
    varargs_handler_err = None
    varargs_plus_handler_counter = 0
    varargs_plus_handler_err = None

    def varargs_handler(*args):
        nonlocal varargs_handler_counter
        nonlocal varargs_handler_err
        try:
            assert args[0] == "hello"
            varargs_handler_counter += 1
        except AssertionError as ex:
            varargs_handler_err = ex

    def varargs_plus_handler(value, *_):
        nonlocal varargs_plus_handler_counter
        nonlocal varargs_plus_handler_err
        try:
            assert value == "hello"
            varargs_plus_handler_counter += 1
        except AssertionError as ex:
            varargs_plus_handler_err = ex

    interface.on_some_signal(varargs_handler)
    interface.on_some_signal(varargs_plus_handler)
    await ping()

    service_interface.SomeSignal()
    await ping()
    assert varargs_handler_err is None
    assert varargs_handler_counter == 1
    assert varargs_plus_handler_err is None
    assert varargs_plus_handler_counter == 1

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_kwargs_callback():
    """Test callback for signal with kwargs."""
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    await bus1.request_name("test.signals.name")
    service_interface = ExampleInterface()
    bus1.export("/test/path", service_interface)

    obj = bus2.get_proxy_object(
        "test.signals.name", "/test/path", bus1._introspect_export_path("/test/path")
    )
    interface = obj.get_interface(service_interface.name)

    async def ping():
        await bus2.call(
            Message(
                destination=bus1.unique_name,
                interface="org.freedesktop.DBus.Peer",
                path="/test/path",
                member="Ping",
            )
        )

    kwargs_handler_counter = 0
    kwargs_handler_err = None
    kwarg_default_handler_counter = 0
    kwarg_default_handler_err = None

    def kwargs_handler(value, **_):
        nonlocal kwargs_handler_counter
        nonlocal kwargs_handler_err
        try:
            assert value == "hello"
            kwargs_handler_counter += 1
        except AssertionError as ex:
            kwargs_handler_err = ex

    def kwarg_default_handler(value, *, _=True):
        nonlocal kwarg_default_handler_counter
        nonlocal kwarg_default_handler_err
        try:
            assert value == "hello"
            kwarg_default_handler_counter += 1
        except AssertionError as ex:
            kwarg_default_handler_err = ex

    interface.on_some_signal(kwargs_handler)
    interface.on_some_signal(kwarg_default_handler)
    await ping()

    service_interface.SomeSignal()
    await ping()
    assert kwargs_handler_err is None
    assert kwargs_handler_counter == 1
    assert kwarg_default_handler_err is None
    assert kwarg_default_handler_counter == 1

    def kwarg_bad_handler(value, *, bad_kwarg):
        pass

    with pytest.raises(TypeError):
        interface.on_some_signal(kwarg_bad_handler)

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_on_signal_type_error():
    """Test on callback raises type errors for invalid callbacks."""
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    await bus1.request_name("test.signals.name")
    service_interface = ExampleInterface()
    bus1.export("/test/path", service_interface)

    obj = bus2.get_proxy_object(
        "test.signals.name", "/test/path", bus1._introspect_export_path("/test/path")
    )
    interface = obj.get_interface(service_interface.name)

    with pytest.raises(TypeError):
        interface.on_some_signal("not_a_callable")

    with pytest.raises(TypeError):
        interface.on_some_signal(lambda a, b: "Too many parameters")

    with pytest.raises(TypeError):
        interface.on_some_signal(lambda: "Too few parameters")

    with pytest.raises(TypeError):
        interface.on_some_signal(lambda a, b, *args: "Too many before varargs")

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_signals_with_changing_owners():
    well_known_name = "test.signals.changing.name"

    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()
    bus3 = await MessageBus().connect()

    async def ping():
        await bus1.call(
            Message(
                destination=bus1.unique_name,
                interface="org.freedesktop.DBus.Peer",
                path="/test/path",
                member="Ping",
            )
        )

    service_interface = ExampleInterface()
    introspection = Node.default()
    introspection.interfaces.append(service_interface.introspect())

    # get the interface before export
    obj = bus1.get_proxy_object(well_known_name, "/test/path", introspection)
    iface = obj.get_interface("test.interface")
    counter = 0

    def handler(what):
        nonlocal counter
        counter += 1

    iface.on_some_signal(handler)
    await ping()

    # now export and get the name
    bus2.export("/test/path", service_interface)
    result = await bus2.request_name(well_known_name)
    assert result is RequestNameReply.PRIMARY_OWNER

    # the signal should work
    service_interface.SomeSignal()
    await ping()
    assert counter == 1
    counter = 0

    # now queue up a transfer of the name
    service_interface2 = ExampleInterface()
    bus3.export("/test/path", service_interface2)
    result = await bus3.request_name(well_known_name)
    assert result is RequestNameReply.IN_QUEUE

    # if it doesn't own the name, the signal shouldn't work here
    service_interface2.SomeSignal()
    await ping()
    assert counter == 0

    # now transfer over the name and it should work
    bus2.disconnect()
    await ping()

    service_interface2.SomeSignal()
    await ping()
    assert counter == 1
    counter = 0

    bus1.disconnect()
    bus2.disconnect()
    bus3.disconnect()



import asyncio

import pytest

from dbus_fast import (
    DBusError,
    ErrorType,
    Message,
    MessageType,
    PropertyAccess,
    Variant,
)
from dbus_fast.aio import MessageBus
from dbus_fast.service import ServiceInterface, dbus_property, method


class ExampleInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)
        self._string_prop = "hi"
        self._readonly_prop = 100
        self._disabled_prop = "1234"
        self._container_prop = [["hello", "world"]]
        self._renamed_prop = "65"

    @dbus_property()
    def string_prop(self) -> "s":
        return self._string_prop

    @string_prop.setter
    def string_prop_setter(self, val: "s"):
        self._string_prop = val

    @dbus_property(PropertyAccess.READ)
    def readonly_prop(self) -> "t":
        return self._readonly_prop

    @dbus_property()
    def container_prop(self) -> "a(ss)":
        return self._container_prop

    @container_prop.setter
    def container_prop(self, val: "a(ss)"):
        self._container_prop = val

    @dbus_property(name="renamed_prop")
    def original_name(self) -> "s":
        return self._renamed_prop

    @original_name.setter
    def original_name_setter(self, val: "s"):
        self._renamed_prop = val

    @dbus_property(disabled=True)
    def disabled_prop(self) -> "s":
        return self._disabled_prop

    @disabled_prop.setter
    def disabled_prop(self, val: "s"):
        self._disabled_prop = val

    @dbus_property(disabled=True)
    def throws_error(self) -> "s":
        raise DBusError("test.error", "told you so")

    @throws_error.setter
    def throws_error(self, val: "s"):
        raise DBusError("test.error", "told you so")

    @dbus_property(PropertyAccess.READ, disabled=True)
    def returns_wrong_type(self) -> "s":
        return 5

    @method()
    def do_emit_properties_changed(self):
        changed = {"string_prop": "asdf"}
        invalidated = ["container_prop"]
        self.emit_properties_changed(changed, invalidated)


class AsyncInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)
        self._string_prop = "hi"
        self._readonly_prop = 100
        self._disabled_prop = "1234"
        self._container_prop = [["hello", "world"]]
        self._renamed_prop = "65"

    @dbus_property()
    async def string_prop(self) -> "s":
        return self._string_prop

    @string_prop.setter
    async def string_prop_setter(self, val: "s"):
        self._string_prop = val

    @dbus_property(PropertyAccess.READ)
    async def readonly_prop(self) -> "t":
        return self._readonly_prop

    @dbus_property()
    async def container_prop(self) -> "a(ss)":
        return self._container_prop

    @container_prop.setter
    async def container_prop(self, val: "a(ss)"):
        self._container_prop = val

    @dbus_property(name="renamed_prop")
    async def original_name(self) -> "s":
        return self._renamed_prop

    @original_name.setter
    async def original_name_setter(self, val: "s"):
        self._renamed_prop = val

    @dbus_property(disabled=True)
    async def disabled_prop(self) -> "s":
        return self._disabled_prop

    @disabled_prop.setter
    async def disabled_prop(self, val: "s"):
        self._disabled_prop = val

    @dbus_property(disabled=True)
    async def throws_error(self) -> "s":
        raise DBusError("test.error", "told you so")

    @throws_error.setter
    async def throws_error(self, val: "s"):
        raise DBusError("test.error", "told you so")

    @dbus_property(PropertyAccess.READ, disabled=True)
    async def returns_wrong_type(self) -> "s":
        return 5

    @method()
    def do_emit_properties_changed(self):
        changed = {"string_prop": "asdf"}
        invalidated = ["container_prop"]
        self.emit_properties_changed(changed, invalidated)


@pytest.mark.parametrize("interface_class", [ExampleInterface, AsyncInterface])
@pytest.mark.asyncio
async def test_property_methods(interface_class):
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = interface_class("test.interface")
    export_path = "/test/path"
    bus1.export(export_path, interface)

    async def call_properties(member, signature, body):
        return await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=export_path,
                interface="org.freedesktop.DBus.Properties",
                member=member,
                signature=signature,
                body=body,
            )
        )

    result = await call_properties("GetAll", "s", [interface.name])

    assert result.message_type == MessageType.METHOD_RETURN, result.body[0]
    assert result.signature == "a{sv}"
    assert result.body == [
        {
            "string_prop": Variant("s", interface._string_prop),
            "readonly_prop": Variant("t", interface._readonly_prop),
            "container_prop": Variant("a(ss)", interface._container_prop),
            "renamed_prop": Variant("s", interface._renamed_prop),
        }
    ]

    result = await call_properties("Get", "ss", [interface.name, "string_prop"])
    assert result.message_type == MessageType.METHOD_RETURN, result.body[0]
    assert result.signature == "v"
    assert result.body == [Variant("s", "hi")]

    result = await call_properties(
        "Set", "ssv", [interface.name, "string_prop", Variant("s", "ho")]
    )
    assert result.message_type == MessageType.METHOD_RETURN, result.body[0]
    assert interface._string_prop == "ho"
    if interface_class is AsyncInterface:
        assert "ho", await interface.string_prop()
    else:
        assert "ho", interface.string_prop

    result = await call_properties(
        "Set", "ssv", [interface.name, "readonly_prop", Variant("t", 100)]
    )
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == ErrorType.PROPERTY_READ_ONLY.value, result.body[0]

    result = await call_properties(
        "Set", "ssv", [interface.name, "disabled_prop", Variant("s", "asdf")]
    )
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == ErrorType.UNKNOWN_PROPERTY.value

    result = await call_properties(
        "Set", "ssv", [interface.name, "not_a_prop", Variant("s", "asdf")]
    )
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == ErrorType.UNKNOWN_PROPERTY.value

    # wrong type
    result = await call_properties(
        "Set", "ssv", [interface.name, "string_prop", Variant("t", 100)]
    )
    assert result.message_type == MessageType.ERROR
    assert result.error_name == ErrorType.INVALID_SIGNATURE.value

    # enable the erroring properties so we can test them
    for prop in ServiceInterface._get_properties(interface):
        if prop.name in ["throws_error", "returns_wrong_type"]:
            prop.disabled = False

    result = await call_properties("Get", "ss", [interface.name, "returns_wrong_type"])
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == ErrorType.SERVICE_ERROR.value

    result = await call_properties(
        "Set", "ssv", [interface.name, "throws_error", Variant("s", "ho")]
    )
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == "test.error"
    assert result.body == ["told you so"]

    result = await call_properties("Get", "ss", [interface.name, "throws_error"])
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == "test.error"
    assert result.body == ["told you so"]

    result = await call_properties("GetAll", "s", [interface.name])
    assert result.message_type == MessageType.ERROR, result.body[0]
    assert result.error_name == "test.error"
    assert result.body == ["told you so"]

    bus1.disconnect()
    bus2.disconnect()
    bus1._sock.close()
    bus2._sock.close()


@pytest.mark.parametrize("interface_class", [ExampleInterface, AsyncInterface])
@pytest.mark.asyncio
async def test_property_changed_signal(interface_class):
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    await bus2.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="AddMatch",
            signature="s",
            body=[f"sender={bus1.unique_name}"],
        )
    )

    interface = interface_class("test.interface")
    export_path = "/test/path"
    bus1.export(export_path, interface)

    async def wait_for_message():
        # TODO timeout
        future = asyncio.get_running_loop().create_future()

        def message_handler(signal):
            if signal.interface == "org.freedesktop.DBus.Properties":
                bus2.remove_message_handler(message_handler)
                future.set_result(signal)

        bus2.add_message_handler(message_handler)
        return await future

    bus2.send(
        Message(
            destination=bus1.unique_name,
            interface=interface.name,
            path=export_path,
            member="do_emit_properties_changed",
        )
    )

    signal = await wait_for_message()
    assert signal.interface == "org.freedesktop.DBus.Properties"
    assert signal.member == "PropertiesChanged"
    assert signal.signature == "sa{sv}as"
    assert signal.body == [
        interface.name,
        {"string_prop": Variant("s", "asdf")},
        ["container_prop"],
    ]

    bus1.disconnect()
    bus2.disconnect()
    bus1._sock.close()
    bus2._sock.close()



import asyncio
import inspect
import logging
import re
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from functools import lru_cache
from typing import Callable, Coroutine, Dict, List, Optional, Type, Union

from . import introspection as intr
from . import message_bus
from ._private.util import replace_idx_with_fds
from .constants import ErrorType, MessageType
from .errors import DBusError, InterfaceNotFoundError
from .message import Message
from .unpack import unpack_variants as unpack
from .validators import assert_bus_name_valid, assert_object_path_valid


@dataclass
class SignalHandler:
    """Signal handler."""

    fn: Callable
    unpack_variants: bool


class BaseProxyInterface:
    """An abstract class representing a proxy to an interface exported on the bus by another client.

    Implementations of this class are not meant to be constructed directly by
    users. Use :func:`BaseProxyObject.get_interface` to get a proxy interface.
    Each message bus implementation provides its own proxy interface
    implementation that will be returned by that method.

    Proxy interfaces can be used to call methods, get properties, and listen to
    signals on the interface. Proxy interfaces are created dynamically with a
    family of methods for each of these operations based on what members the
    interface exposes. Each proxy interface implementation exposes these
    members in a different way depending on the features of the backend. See
    the documentation of the proxy interface implementation you use for more
    details.

    :ivar bus_name: The name of the bus this interface is exported on.
    :vartype bus_name: str
    :ivar path: The object path exported on the client that owns the bus name.
    :vartype path: str
    :ivar introspection: Parsed introspection data for the proxy interface.
    :vartype introspection: :class:`Node <dbus_fast.introspection.Interface>`
    :ivar bus: The message bus this proxy interface is connected to.
    :vartype bus: :class:`BaseMessageBus <dbus_fast.message_bus.BaseMessageBus>`
    """

    def __init__(
        self,
        bus_name: str,
        path: str,
        introspection: intr.Interface,
        bus: "message_bus.BaseMessageBus",
    ) -> None:
        self.bus_name = bus_name
        self.path = path
        self.introspection = introspection
        self.bus = bus
        self._signal_handlers: Dict[str, List[SignalHandler]] = {}
        self._signal_match_rule = f"type='signal',sender={bus_name},interface={introspection.name},path={path}"

    _underscorer1 = re.compile(r"(.)([A-Z][a-z]+)")
    _underscorer2 = re.compile(r"([a-z0-9])([A-Z])")

    @staticmethod
    @lru_cache(maxsize=128)
    def _to_snake_case(member: str) -> str:
        subbed = BaseProxyInterface._underscorer1.sub(r"\1_\2", member)
        return BaseProxyInterface._underscorer2.sub(r"\1_\2", subbed).lower()

    @staticmethod
    def _check_method_return(msg: Message, signature: Optional[str] = None):
        if msg.message_type == MessageType.ERROR:
            raise DBusError._from_message(msg)
        elif msg.message_type != MessageType.METHOD_RETURN:
            raise DBusError(
                ErrorType.CLIENT_ERROR, "method call didnt return a method return", msg
            )
        elif signature is not None and msg.signature != signature:
            raise DBusError(
                ErrorType.CLIENT_ERROR,
                f'method call returned unexpected signature: "{msg.signature}"',
                msg,
            )

    def _add_method(self, intr_method: intr.Method) -> None:
        raise NotImplementedError("this must be implemented in the inheriting class")

    def _add_property(self, intr_property: intr.Property) -> None:
        raise NotImplementedError("this must be implemented in the inheriting class")

    def _message_handler(self, msg: Message) -> None:
        if (
            msg.message_type != MessageType.SIGNAL
            or msg.interface != self.introspection.name
            or msg.path != self.path
            or msg.member not in self._signal_handlers
        ):
            return

        if (
            msg.sender != self.bus_name
            and self.bus._name_owners.get(self.bus_name, "") != msg.sender
        ):
            # The sender is always a unique name, but the bus name given might
            # be a well known name. If the sender isn't an exact match, check
            # to see if it owns the bus_name we were given from the cache kept
            # on the bus for this purpose.
            return

        match = [s for s in self.introspection.signals if s.name == msg.member]
        if not len(match):
            return
        intr_signal = match[0]
        if intr_signal.signature != msg.signature:
            logging.warning(
                f'got signal "{self.introspection.name}.{msg.member}" with unexpected signature "{msg.signature}"'
            )
            return

        body = replace_idx_with_fds(msg.signature, msg.body, msg.unix_fds)
        no_sig = None
        for handler in self._signal_handlers[msg.member]:
            if handler.unpack_variants:
                if not no_sig:
                    no_sig = unpack(body)
                data = no_sig
            else:
                data = body

            cb_result = handler.fn(*data)
            if isinstance(cb_result, Coroutine):
                asyncio.create_task(cb_result)

    def _add_signal(self, intr_signal: intr.Signal, interface: intr.Interface) -> None:
        def on_signal_fn(fn: Callable, *, unpack_variants: bool = False):
            fn_signature = inspect.signature(fn)
            if 0 < len(
                [
                    par
                    for par in fn_signature.parameters.values()
                    if par.kind == inspect.Parameter.KEYWORD_ONLY
                    and par.default == inspect.Parameter.empty
                ]
            ):
                raise TypeError(
                    "reply_notify cannot have required keyword only parameters"
                )

            positional_params = [
                par.kind
                for par in fn_signature.parameters.values()
                if par.kind
                not in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD]
            ]
            if len(positional_params) != len(intr_signal.args) and (
                inspect.Parameter.VAR_POSITIONAL not in positional_params
                or len(positional_params) - 1 > len(intr_signal.args)
            ):
                raise TypeError(
                    f"reply_notify must be a function with {len(intr_signal.args)} positional parameters"
                )

            if not self._signal_handlers:
                self.bus._add_match_rule(self._signal_match_rule)
                self.bus.add_message_handler(self._message_handler)

            if intr_signal.name not in self._signal_handlers:
                self._signal_handlers[intr_signal.name] = []

            self._signal_handlers[intr_signal.name].append(
                SignalHandler(fn, unpack_variants)
            )

        def off_signal_fn(fn: Callable, *, unpack_variants: bool = False) -> None:
            try:
                i = self._signal_handlers[intr_signal.name].index(
                    SignalHandler(fn, unpack_variants)
                )
                del self._signal_handlers[intr_signal.name][i]
                if not self._signal_handlers[intr_signal.name]:
                    del self._signal_handlers[intr_signal.name]
            except (KeyError, ValueError):
                return

            if not self._signal_handlers:
                self.bus._remove_match_rule(self._signal_match_rule)
                self.bus.remove_message_handler(self._message_handler)

        snake_case = BaseProxyInterface._to_snake_case(intr_signal.name)
        setattr(interface, f"on_{snake_case}", on_signal_fn)
        setattr(interface, f"off_{snake_case}", off_signal_fn)


class BaseProxyObject:
    """An abstract class representing a proxy to an object exported on the bus by another client.

    Implementations of this class are not meant to be constructed directly. Use
    :func:`BaseMessageBus.get_proxy_object()
    <dbus_fast.message_bus.BaseMessageBus.get_proxy_object>` to get a proxy
    object. Each message bus implementation provides its own proxy object
    implementation that will be returned by that method.

    The primary use of the proxy object is to select a proxy interface to act
    on. Information on what interfaces are available is provided by
    introspection data provided to this class. This introspection data can
    either be included in your project as an XML file (recommended) or
    retrieved from the ``org.freedesktop.DBus.Introspectable`` interface at
    runtime.

    :ivar bus_name: The name of the bus this object is exported on.
    :vartype bus_name: str
    :ivar path: The object path exported on the client that owns the bus name.
    :vartype path: str
    :ivar introspection: Parsed introspection data for the proxy object.
    :vartype introspection: :class:`Node <dbus_fast.introspection.Node>`
    :ivar bus: The message bus this proxy object is connected to.
    :vartype bus: :class:`BaseMessageBus <dbus_fast.message_bus.BaseMessageBus>`
    :ivar ~.ProxyInterface: The proxy interface class this proxy object uses.
    :vartype ~.ProxyInterface: Type[:class:`BaseProxyInterface <dbus_fast.proxy_object.BaseProxyObject>`]
    :ivar child_paths: A list of absolute object paths of the children of this object.
    :vartype child_paths: list(str)

    :raises:
        - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If the given bus name is not valid.
        - :class:`InvalidObjectPathError <dbus_fast.InvalidObjectPathError>` - If the given object path is not valid.
        - :class:`InvalidIntrospectionError <dbus_fast.InvalidIntrospectionError>` - If the introspection data for the node is not valid.
    """

    def __init__(
        self,
        bus_name: str,
        path: str,
        introspection: Union[intr.Node, str, ET.Element],
        bus: "message_bus.BaseMessageBus",
        ProxyInterface: Type[BaseProxyInterface],
    ) -> None:
        assert_object_path_valid(path)
        assert_bus_name_valid(bus_name)

        if not isinstance(bus, message_bus.BaseMessageBus):
            raise TypeError("bus must be an instance of BaseMessageBus")
        if not issubclass(ProxyInterface, BaseProxyInterface):
            raise TypeError("ProxyInterface must be an instance of BaseProxyInterface")

        if type(introspection) is intr.Node:
            self.introspection = introspection
        elif type(introspection) is str:
            self.introspection = intr.Node.parse(introspection)
        elif type(introspection) is ET.Element:
            self.introspection = intr.Node.from_xml(introspection)
        else:
            raise TypeError(
                "introspection must be xml node introspection or introspection.Node class"
            )

        self.bus_name = bus_name
        self.path = path
        self.bus = bus
        self.ProxyInterface = ProxyInterface
        self.child_paths = [f"{path}/{n.name}" for n in self.introspection.nodes]

        self._interfaces = {}

        # lazy loaded by get_children()
        self._children = None

    def get_interface(self, name: str) -> BaseProxyInterface:
        """Get an interface exported on this proxy object and connect it to the bus.

        :param name: The name of the interface to retrieve.
        :type name: str

        :raises:
            - :class:`InterfaceNotFoundError <dbus_fast.InterfaceNotFoundError>` - If there is no interface by this name exported on the bus.
        """
        if name in self._interfaces:
            return self._interfaces[name]

        try:
            intr_interface = next(
                i for i in self.introspection.interfaces if i.name == name
            )
        except StopIteration:
            raise InterfaceNotFoundError(f"interface not found on this object: {name}")

        interface = self.ProxyInterface(
            self.bus_name, self.path, intr_interface, self.bus
        )

        for intr_method in intr_interface.methods:
            interface._add_method(intr_method)
        for intr_property in intr_interface.properties:
            interface._add_property(intr_property)
        for intr_signal in intr_interface.signals:
            interface._add_signal(intr_signal, interface)

        def get_owner_notify(msg: Message, err: Optional[Exception]) -> None:
            if err:
                logging.error(f'getting name owner for "{name}" failed, {err}')
                return
            if msg.message_type == MessageType.ERROR:
                if msg.error_name != ErrorType.NAME_HAS_NO_OWNER.value:
                    logging.error(
                        f'getting name owner for "{name}" failed, {msg.body[0]}'
                    )
                return

            self.bus._name_owners[self.bus_name] = msg.body[0]

        if self.bus_name[0] != ":" and not self.bus._name_owners.get(self.bus_name, ""):
            self.bus._call(
                Message(
                    destination="org.freedesktop.DBus",
                    interface="org.freedesktop.DBus",
                    path="/org/freedesktop/DBus",
                    member="GetNameOwner",
                    signature="s",
                    body=[self.bus_name],
                ),
                get_owner_notify,
            )

        self._interfaces[name] = interface
        return interface

    def get_children(self) -> List["BaseProxyObject"]:
        """Get the child nodes of this proxy object according to the introspection data."""
        if self._children is None:
            self._children = [
                self.__class__(self.bus_name, self.path, child, self.bus)
                for child in self.introspection.nodes
            ]

        return self._children



"""Build optional cython modules."""

import os
from distutils.command.build_ext import build_ext


class BuildExt(build_ext):
    def build_extensions(self):
        try:
            super().build_extensions()
        except Exception:
            pass


def build(setup_kwargs):
    if os.environ.get("SKIP_CYTHON", False):
        return
    try:
        from Cython.Build import cythonize

        setup_kwargs.update(
            dict(
                ext_modules=cythonize(
                    [
                        "src/dbus_fast/aio/message_reader.py",
                        "src/dbus_fast/message.py",
                        "src/dbus_fast/message_bus.py",
                        "src/dbus_fast/service.py",
                        "src/dbus_fast/signature.py",
                        "src/dbus_fast/unpack.py",
                        "src/dbus_fast/_private/address.py",
                        "src/dbus_fast/_private/marshaller.py",
                        "src/dbus_fast/_private/unmarshaller.py",
                    ],
                    compiler_directives={"language_level": "3"},  # Python 3
                ),
                cmdclass=dict(build_ext=BuildExt),
            )
        )
        setup_kwargs["exclude_package_data"] = {
            pkg: ["*.c"] for pkg in setup_kwargs["packages"]
        }
    except Exception:
        if os.environ.get("REQUIRE_CYTHON"):
            raise
        pass



"""cdefs for service.py"""

import cython

from .message cimport Message
from .signature cimport SignatureTree


cdef class _Method:

    cdef public str name
    cdef public object fn
    cdef public bint disabled
    cdef public object introspection
    cdef public str in_signature
    cdef public str out_signature
    cdef public SignatureTree in_signature_tree
    cdef public SignatureTree out_signature_tree



cdef tuple _real_fn_result_to_body(
    object result,
    SignatureTree signature_tree,
    bint replace_fds
)

cdef class ServiceInterface:

    cdef public str name
    cdef list __methods
    cdef list __properties
    cdef list __signals
    cdef set __buses
    cdef dict __handlers

    @staticmethod
    cdef list _c_get_methods(ServiceInterface interface)

    @staticmethod
    cdef object _c_get_handler(ServiceInterface interface, _Method method, object bus)

    @staticmethod
    cdef list _c_msg_body_to_args(Message msg)

    @staticmethod
    cdef tuple _c_fn_result_to_body(
        object result,
        SignatureTree signature_tree,
        bint replace_fds,
    )



import pytest

from dbus_fast import Message, MessageType
from dbus_fast import introspection as intr
from dbus_fast.aio import MessageBus
from dbus_fast.constants import ErrorType
from dbus_fast.service import PropertyAccess, ServiceInterface, dbus_property
from dbus_fast.signature import Variant

standard_interfaces_count = len(intr.Node.default().interfaces)


class ExampleInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)


class ExampleComplexInterface(ServiceInterface):
    def __init__(self, name):
        self._foo = 42
        self._bar = "str"
        self._async_prop = "async"
        super().__init__(name)

    @dbus_property(access=PropertyAccess.READ)
    def Foo(self) -> "y":
        return self._foo

    @dbus_property(access=PropertyAccess.READ)
    def Bar(self) -> "s":
        return self._bar

    @dbus_property(access=PropertyAccess.READ)
    async def AsyncProp(self) -> "s":
        return self._async_prop


@pytest.mark.asyncio
async def test_introspectable_interface():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = ExampleInterface("test.interface")
    interface2 = ExampleInterface("test.interface2")

    export_path = "/test/path"
    bus1.export(export_path, interface)
    bus1.export(export_path, interface2)

    reply = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path=export_path,
            interface="org.freedesktop.DBus.Introspectable",
            member="Introspect",
        )
    )

    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "s"
    node = intr.Node.parse(reply.body[0])
    assert len(node.interfaces) == standard_interfaces_count + 2
    assert node.interfaces[-1].name == "test.interface2"
    assert node.interfaces[-2].name == "test.interface"
    assert not node.nodes

    # introspect works on every path
    reply = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path="/path/doesnt/exist",
            interface="org.freedesktop.DBus.Introspectable",
            member="Introspect",
        )
    )
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "s"
    node = intr.Node.parse(reply.body[0])
    assert not node.interfaces
    assert not node.nodes

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_introspect_matching_sub_paths():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = ExampleInterface("test.interface1")

    bus1.export("/a/test/path1", interface)
    bus1.export("/a/subpath/a/test/path2", interface)

    async def introspect_subpath(path, expected_subnodes):
        reply = await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=path,
                interface="org.freedesktop.DBus.Introspectable",
                member="Introspect",
            )
        )
        assert reply.signature == "s"
        node = intr.Node.parse(reply.body[0])
        assert {n.name for n in node.nodes} == expected_subnodes

    await introspect_subpath("/", {"a"})
    await introspect_subpath("/a", {"test", "subpath"})
    await introspect_subpath("/a/test", {"path1"})
    await introspect_subpath("/a/test/path1", set())
    await introspect_subpath("/a/subpath/a/test", {"path2"})
    await introspect_subpath("/a/subpath/a/test/path2", set())

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_peer_interface():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    reply = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path="/path/doesnt/exist",
            interface="org.freedesktop.DBus.Peer",
            member="Ping",
        )
    )

    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == ""

    reply = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path="/path/doesnt/exist",
            interface="org.freedesktop.DBus.Peer",
            member="GetMachineId",
            signature="",
        )
    )

    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "s"

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_object_manager():
    expected_reply = {
        "/test/path/deeper": {
            "test.interface2": {
                "Bar": Variant("s", "str"),
                "Foo": Variant("y", 42),
                "AsyncProp": Variant("s", "async"),
            }
        }
    }
    reply_ext = {
        "/test/path": {
            "test.interface1": {},
            "test.interface2": {
                "Bar": Variant("s", "str"),
                "Foo": Variant("y", 42),
                "AsyncProp": Variant("s", "async"),
            },
        }
    }

    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = ExampleInterface("test.interface1")
    interface2 = ExampleComplexInterface("test.interface2")

    export_path = "/test/path"
    bus1.export(export_path, interface)
    bus1.export(export_path, interface2)
    bus1.export(export_path + "/deeper", interface2)

    reply_root = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path="/",
            interface="org.freedesktop.DBus.ObjectManager",
            member="GetManagedObjects",
        )
    )

    reply_level1 = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path=export_path,
            interface="org.freedesktop.DBus.ObjectManager",
            member="GetManagedObjects",
        )
    )

    reply_level2 = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path=export_path + "/deeper",
            interface="org.freedesktop.DBus.ObjectManager",
            member="GetManagedObjects",
        )
    )

    assert reply_root.signature == "a{oa{sa{sv}}}"
    assert reply_level1.signature == "a{oa{sa{sv}}}"
    assert reply_level2.signature == "a{oa{sa{sv}}}"

    assert reply_level2.body == [{}]
    assert reply_level1.body == [expected_reply]
    expected_reply.update(reply_ext)
    assert reply_root.body == [expected_reply]

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_standard_interface_properties():
    # standard interfaces have no properties, but should still behave correctly
    # when you try to call the methods anyway (#49)
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()
    interface = ExampleInterface("test.interface1")
    export_path = "/test/path"
    bus1.export(export_path, interface)

    for iface in [
        "org.freedesktop.DBus.Properties",
        "org.freedesktop.DBus.Introspectable",
        "org.freedesktop.DBus.Peer",
        "org.freedesktop.DBus.ObjectManager",
    ]:
        result = await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=export_path,
                interface="org.freedesktop.DBus.Properties",
                member="Get",
                signature="ss",
                body=[iface, "anything"],
            )
        )
        assert result.message_type is MessageType.ERROR
        assert result.error_name == ErrorType.UNKNOWN_PROPERTY.value

        result = await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=export_path,
                interface="org.freedesktop.DBus.Properties",
                member="Set",
                signature="ssv",
                body=[iface, "anything", Variant("s", "new thing")],
            )
        )
        assert result.message_type is MessageType.ERROR
        assert result.error_name == ErrorType.UNKNOWN_PROPERTY.value

        result = await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=export_path,
                interface="org.freedesktop.DBus.Properties",
                member="GetAll",
                signature="s",
                body=[iface],
            )
        )
        assert result.message_type is MessageType.METHOD_RETURN
        assert result.body == [{}]

    bus1.disconnect()
    bus2.disconnect()



import pytest

from dbus_fast import Message, MessageFlag, MessageType
from dbus_fast.aio import MessageBus


@pytest.mark.asyncio
async def test_standard_interfaces():
    bus = await MessageBus().connect()
    msg = Message(
        destination="org.freedesktop.DBus",
        path="/org/freedesktop/DBus",
        interface="org.freedesktop.DBus",
        member="ListNames",
        serial=bus.next_serial(),
    )
    reply = await bus.call(msg)

    assert reply.message_type == MessageType.METHOD_RETURN
    assert reply.reply_serial == msg.serial
    assert reply.signature == "as"
    assert bus.unique_name in reply.body[0]

    msg.interface = "org.freedesktop.DBus.Introspectable"
    msg.member = "Introspect"
    msg.serial = bus.next_serial()

    reply = await bus.call(msg)
    assert reply.message_type == MessageType.METHOD_RETURN
    assert reply.reply_serial == msg.serial
    assert reply.signature == "s"
    assert type(reply.body[0]) is str

    msg.member = "MemberDoesNotExist"
    msg.serial = bus.next_serial()

    reply = await bus.call(msg)
    assert reply.message_type == MessageType.ERROR
    assert reply.reply_serial == msg.serial
    assert reply.error_name
    assert reply.signature == "s"
    assert type(reply.body[0]) is str

    bus.disconnect()


@pytest.mark.asyncio
async def test_sending_messages_between_buses():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    msg = Message(
        destination=bus1.unique_name,
        path="/org/test/path",
        interface="org.test.iface",
        member="SomeMember",
        serial=bus2.next_serial(),
    )

    def message_handler(sent):
        if sent.sender == bus2.unique_name and sent.serial == msg.serial:
            assert sent.path == msg.path
            assert sent.serial == msg.serial
            assert sent.interface == msg.interface
            assert sent.member == msg.member
            bus1.send(Message.new_method_return(sent, "s", ["got it"]))
            bus1.remove_message_handler(message_handler)
            return True

    bus1.add_message_handler(message_handler)

    reply = await bus2.call(msg)

    assert reply.message_type == MessageType.METHOD_RETURN
    assert reply.sender == bus1.unique_name
    assert reply.signature == "s"
    assert reply.body == ["got it"]
    assert reply.reply_serial == msg.serial

    def message_handler_error(sent):
        if sent.sender == bus2.unique_name and sent.serial == msg.serial:
            assert sent.path == msg.path
            assert sent.serial == msg.serial
            assert sent.interface == msg.interface
            assert sent.member == msg.member
            bus1.send(Message.new_error(sent, "org.test.Error", "throwing an error"))
            bus1.remove_message_handler(message_handler_error)
            return True

    bus1.add_message_handler(message_handler_error)

    msg.serial = bus2.next_serial()

    reply = await bus2.call(msg)

    assert reply.message_type == MessageType.ERROR
    assert reply.sender == bus1.unique_name
    assert reply.reply_serial == msg.serial
    assert reply.error_name == "org.test.Error"
    assert reply.signature == "s"
    assert reply.body == ["throwing an error"]

    msg.serial = bus2.next_serial()
    msg.flags = MessageFlag.NO_REPLY_EXPECTED
    reply = await bus2.call(msg)
    assert reply is None

    bus1.disconnect()
    bus2.disconnect()


@pytest.mark.asyncio
async def test_sending_signals_between_buses(event_loop):
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    add_match_msg = Message(
        destination="org.freedesktop.DBus",
        path="/org/freedesktop/DBus",
        interface="org.freedesktop.DBus",
        member="AddMatch",
        signature="s",
        body=[f"sender={bus2.unique_name}"],
    )

    await bus1.call(add_match_msg)

    async def wait_for_message():
        future = event_loop.create_future()

        def message_handler(signal):
            if signal.sender == bus2.unique_name:
                bus1.remove_message_handler(message_handler)
                future.set_result(signal)

        bus1.add_message_handler(message_handler)
        return await future

    bus2.send(
        Message.new_signal(
            "/org/test/path", "org.test.interface", "SomeSignal", "s", ["a signal"]
        )
    )

    signal = await wait_for_message()

    assert signal.message_type == MessageType.SIGNAL
    assert signal.path == "/org/test/path"
    assert signal.interface == "org.test.interface"
    assert signal.member == "SomeSignal"
    assert signal.signature == "s"
    assert signal.body == ["a signal"]

    bus1.disconnect()
    bus2.disconnect()



from dbus_fast import PropertyAccess
from dbus_fast import introspection as intr
from dbus_fast.service import ServiceInterface, dbus_property, method, signal


class ExampleInterface(ServiceInterface):
    def __init__(self):
        super().__init__("test.interface")
        self._some_prop = 55
        self._another_prop = 101
        self._weird_prop = 500

    @method()
    def some_method(self, one: "s", two: "s") -> "s":
        return "hello"

    @method(name="renamed_method", disabled=True)
    def another_method(self, eight: "o", six: "t"):
        pass

    @signal()
    def some_signal(self) -> "as":
        return ["result"]

    @signal(name="renamed_signal", disabled=True)
    def another_signal(self) -> "(dodo)":
        return [1, "/", 1, "/"]

    @dbus_property(
        name="renamed_readonly_property", access=PropertyAccess.READ, disabled=True
    )
    def another_prop(self) -> "t":
        return self._another_prop

    @dbus_property()
    def some_prop(self) -> "u":
        return self._some_prop

    @some_prop.setter
    def some_prop(self, val: "u"):
        self._some_prop = val + 1

    # for this one, the setter has a different name than the getter which is a
    # special case in the code
    @dbus_property()
    def weird_prop(self) -> "t":
        return self._weird_prop

    @weird_prop.setter
    def setter_for_weird_prop(self, val: "t"):
        self._weird_prop = val


def test_method_decorator():
    interface = ExampleInterface()
    assert interface.name == "test.interface"

    properties = ServiceInterface._get_properties(interface)
    methods = ServiceInterface._get_methods(interface)
    signals = ServiceInterface._get_signals(interface)

    assert len(methods) == 2

    method = methods[0]
    assert method.name == "renamed_method"
    assert method.in_signature == "ot"
    assert method.out_signature == ""
    assert method.disabled
    assert type(method.introspection) is intr.Method

    method = methods[1]
    assert method.name == "some_method"
    assert method.in_signature == "ss"
    assert method.out_signature == "s"
    assert not method.disabled
    assert type(method.introspection) is intr.Method

    assert len(signals) == 2

    signal = signals[0]
    assert signal.name == "renamed_signal"
    assert signal.signature == "(dodo)"
    assert signal.disabled
    assert type(signal.introspection) is intr.Signal

    signal = signals[1]
    assert signal.name == "some_signal"
    assert signal.signature == "as"
    assert not signal.disabled
    assert type(signal.introspection) is intr.Signal

    assert len(properties) == 3

    renamed_readonly_prop = properties[0]
    assert renamed_readonly_prop.name == "renamed_readonly_property"
    assert renamed_readonly_prop.signature == "t"
    assert renamed_readonly_prop.access == PropertyAccess.READ
    assert renamed_readonly_prop.disabled
    assert type(renamed_readonly_prop.introspection) is intr.Property

    weird_prop = properties[1]
    assert weird_prop.name == "weird_prop"
    assert weird_prop.access == PropertyAccess.READWRITE
    assert weird_prop.signature == "t"
    assert not weird_prop.disabled
    assert weird_prop.prop_getter is not None
    assert weird_prop.prop_getter.__name__ == "weird_prop"
    assert weird_prop.prop_setter is not None
    assert weird_prop.prop_setter.__name__ == "setter_for_weird_prop"
    assert type(weird_prop.introspection) is intr.Property

    prop = properties[2]
    assert prop.name == "some_prop"
    assert prop.access == PropertyAccess.READWRITE
    assert prop.signature == "u"
    assert not prop.disabled
    assert prop.prop_getter is not None
    assert prop.prop_setter is not None
    assert type(prop.introspection) is intr.Property

    # make sure the getter and setter actually work
    assert interface._some_prop == 55
    interface._some_prop = 555
    assert interface.some_prop == 555

    assert interface._weird_prop == 500
    assert weird_prop.prop_getter(interface) == 500
    interface._weird_prop = 1001
    assert interface._weird_prop == 1001
    weird_prop.prop_setter(interface, 600)
    assert interface._weird_prop == 600


def test_interface_introspection():
    interface = ExampleInterface()
    intr_interface = interface.introspect()
    assert type(intr_interface) is intr.Interface

    xml = intr_interface.to_xml()

    assert xml.tag == "interface"
    assert xml.attrib.get("name", None) == "test.interface"

    methods = xml.findall("method")
    signals = xml.findall("signal")
    properties = xml.findall("property")

    assert len(xml) == 4
    assert len(methods) == 1
    assert len(signals) == 1
    assert len(properties) == 2



import pytest

from dbus_fast import (
    DBusError,
    ErrorType,
    Message,
    MessageFlag,
    MessageType,
    SignatureTree,
    Variant,
)
from dbus_fast.aio import MessageBus
from dbus_fast.service import ServiceInterface, method


class ExampleInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)

    @method()
    def echo(self, what: "s") -> "s":
        assert type(self) is ExampleInterface
        return what

    @method()
    def echo_multiple(self, what1: "s", what2: "s") -> "ss":
        assert type(self) is ExampleInterface
        return [what1, what2]

    @method()
    def echo_containers(
        self, array: "as", variant: "v", dict_entries: "a{sv}", struct: "(s(s(v)))"
    ) -> "asva{sv}(s(s(v)))":
        assert type(self) is ExampleInterface
        return [array, variant, dict_entries, struct]

    @method()
    def ping(self):
        assert type(self) is ExampleInterface
        pass

    @method(name="renamed")
    def original_name(self):
        assert type(self) is ExampleInterface
        pass

    @method(disabled=True)
    def not_here(self):
        assert type(self) is ExampleInterface
        pass

    @method()
    def throws_unexpected_error(self):
        assert type(self) is ExampleInterface
        raise Exception("oops")

    @method()
    def throws_dbus_error(self):
        assert type(self) is ExampleInterface
        raise DBusError("test.error", "an error ocurred")


class AsyncInterface(ServiceInterface):
    def __init__(self, name):
        super().__init__(name)

    @method()
    async def echo(self, what: "s") -> "s":
        assert type(self) is AsyncInterface
        return what

    @method()
    async def echo_multiple(self, what1: "s", what2: "s") -> "ss":
        assert type(self) is AsyncInterface
        return [what1, what2]

    @method()
    async def echo_containers(
        self, array: "as", variant: "v", dict_entries: "a{sv}", struct: "(s(s(v)))"
    ) -> "asva{sv}(s(s(v)))":
        assert type(self) is AsyncInterface
        return [array, variant, dict_entries, struct]

    @method()
    async def ping(self):
        assert type(self) is AsyncInterface
        pass

    @method(name="renamed")
    async def original_name(self):
        assert type(self) is AsyncInterface
        pass

    @method(disabled=True)
    async def not_here(self):
        assert type(self) is AsyncInterface
        pass

    @method()
    async def throws_unexpected_error(self):
        assert type(self) is AsyncInterface
        raise Exception("oops")

    @method()
    def throws_dbus_error(self):
        assert type(self) is AsyncInterface
        raise DBusError("test.error", "an error ocurred")


@pytest.mark.parametrize("interface_class", [ExampleInterface, AsyncInterface])
@pytest.mark.asyncio
async def test_methods(interface_class):
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = interface_class("test.interface")
    export_path = "/test/path"

    async def call(member, signature="", body=[], flags=MessageFlag.NONE):
        return await bus2.call(
            Message(
                destination=bus1.unique_name,
                path=export_path,
                interface=interface.name,
                member=member,
                signature=signature,
                body=body,
                flags=flags,
            )
        )

    bus1.export(export_path, interface)

    body = ["hello world"]
    reply = await call("echo", "s", body)

    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "s"
    assert reply.body == body

    body = ["hello", "world"]
    reply = await call("echo_multiple", "ss", body)
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "ss"
    assert reply.body == body

    body = [
        ["hello", "world"],
        Variant("v", Variant("(ss)", ["hello", "world"])),
        {"foo": Variant("t", 100)},
        ["one", ["two", [Variant("s", "three")]]],
    ]
    signature = "asva{sv}(s(s(v)))"
    SignatureTree(signature).verify(body)
    reply = await call("echo_containers", signature, body)
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == signature
    assert reply.body == body

    reply = await call("ping")
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == ""
    assert reply.body == []

    reply = await call("throws_unexpected_error")
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == ErrorType.SERVICE_ERROR.value, reply.body[0]

    reply = await call("throws_dbus_error")
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == "test.error", reply.body[0]
    assert reply.body == ["an error ocurred"]

    reply = await call("ping", flags=MessageFlag.NO_REPLY_EXPECTED)
    assert reply is None

    reply = await call("throws_unexpected_error", flags=MessageFlag.NO_REPLY_EXPECTED)
    assert reply is None

    reply = await call("throws_dbus_error", flags=MessageFlag.NO_REPLY_EXPECTED)
    assert reply is None

    bus1.disconnect()
    bus2.disconnect()
    bus1._sock.close()
    bus2._sock.close()



from . import introspection, message_bus, proxy_object, service
from .constants import (
    ArgDirection,
    BusType,
    ErrorType,
    MessageFlag,
    MessageType,
    NameFlag,
    PropertyAccess,
    ReleaseNameReply,
    RequestNameReply,
)
from .errors import (
    AuthError,
    DBusError,
    InterfaceNotFoundError,
    InvalidAddressError,
    InvalidBusNameError,
    InvalidInterfaceNameError,
    InvalidIntrospectionError,
    InvalidMemberNameError,
    InvalidMessageError,
    InvalidObjectPathError,
    InvalidSignatureError,
    SignalDisabledError,
    SignatureBodyMismatchError,
)
from .message import Message
from .signature import SignatureTree, SignatureType, Variant
from .unpack import unpack_variants
from .validators import (
    assert_bus_name_valid,
    assert_interface_name_valid,
    assert_member_name_valid,
    assert_object_path_valid,
    is_bus_name_valid,
    is_interface_name_valid,
    is_member_name_valid,
    is_object_path_valid,
)

__all__ = [
    "introspection",
    "message_bus",
    "proxy_object",
    "service",
    "ArgDirection",
    "BusType",
    "ErrorType",
    "MessageFlag",
    "MessageType",
    "NameFlag",
    "PropertyAccess",
    "ReleaseNameReply",
    "RequestNameReply",
    "AuthError",
    "DBusError",
    "InterfaceNotFoundError",
    "InvalidAddressError",
    "InvalidBusNameError",
    "InvalidInterfaceNameError",
    "InvalidIntrospectionError",
    "InvalidMemberNameError",
    "InvalidMessageError",
    "InvalidObjectPathError",
    "InvalidSignatureError",
    "SignalDisabledError",
    "SignatureBodyMismatchError",
    "Message",
    "SignatureTree",
    "SignatureType",
    "Variant",
    "assert_bus_name_valid",
    "assert_interface_name_valid",
    "assert_member_name_valid",
    "assert_object_path_valid",
    "is_bus_name_valid",
    "is_interface_name_valid",
    "is_member_name_valid",
    "is_object_path_valid",
    "unpack_variants",
]



import traceback
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type

from .constants import ErrorType
from .errors import DBusError
from .message import Message

if TYPE_CHECKING:
    from .message_bus import BaseMessageBus


class SendReply:
    """A context manager to send a reply to a message."""

    __slots__ = ("_bus", "_msg")

    def __init__(self, bus: "BaseMessageBus", msg: Message) -> None:
        """Create a new reply context manager."""
        self._bus = bus
        self._msg = msg

    def __enter__(self):
        return self

    def __call__(self, reply: Message) -> None:
        self._bus.send(reply)

    def _exit(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        tb: Optional[TracebackType],
    ) -> bool:
        if exc_value:
            if isinstance(exc_value, DBusError):
                self(exc_value._as_message(self._msg))
            else:
                self(
                    Message.new_error(
                        self._msg,
                        ErrorType.SERVICE_ERROR,
                        f"The service interface raised an error: {exc_value}.\n{traceback.format_tb(tb)}",
                    )
                )
            return True

        return False

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        tb: Optional[TracebackType],
    ) -> bool:
        return self._exit(exc_type, exc_value, tb)

    def send_error(self, exc: Exception) -> None:
        self._exit(exc.__class__, exc, exc.__traceback__)



from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .errors import InvalidSignatureError, SignatureBodyMismatchError
from .validators import is_object_path_valid


class SignatureType:
    """A class that represents a single complete type within a signature.

    This class is not meant to be constructed directly. Use the :class:`SignatureTree`
    class to parse signatures.

    :ivar ~.signature: The signature of this complete type.
    :vartype ~.signature: str

    :ivar children: A list of child types if this is a container type. Arrays \
    have one child type, dict entries have two child types (key and value), and \
    structs have child types equal to the number of struct members.
    :vartype children: list(:class:`SignatureType`)
    """

    _tokens = "ybnqiuxtdsogavh({"
    __slots__ = ("token", "children", "_signature")

    def __init__(self, token: str) -> None:
        """Init a new SignatureType."""
        self.token: str = token
        self.children: List[SignatureType] = []
        self._signature: Optional[str] = None

    def __eq__(self, other: Any) -> bool:
        """Compare this type to another type or signature string."""
        if type(other) is SignatureType:
            return self.signature == other.signature
        return super().__eq__(other)

    def _collapse(self) -> str:
        """Collapse this type into a signature string."""
        if self.token not in "a({":
            return self.token

        signature = [self.token]

        for child in self.children:
            signature.append(child._collapse())

        if self.token == "(":
            signature.append(")")
        elif self.token == "{":
            signature.append("}")

        return "".join(signature)

    @property
    def signature(self) -> str:
        if self._signature is not None:
            return self._signature
        self._signature = self._collapse()
        return self._signature

    @staticmethod
    def _parse_next(signature: str) -> Tuple["SignatureType", str]:
        if not signature:
            raise InvalidSignatureError("Cannot parse an empty signature")

        token = signature[0]

        if token not in SignatureType._tokens:
            raise InvalidSignatureError(f'got unexpected token: "{token}"')

        # container types
        if token == "a":
            self = SignatureType("a")
            (child, signature) = SignatureType._parse_next(signature[1:])
            if not child:
                raise InvalidSignatureError("missing type for array")
            self.children.append(child)
            return (self, signature)
        elif token == "(":
            self = SignatureType("(")
            signature = signature[1:]
            while True:
                (child, signature) = SignatureType._parse_next(signature)
                if not signature:
                    raise InvalidSignatureError('missing closing ")" for struct')
                self.children.append(child)
                if signature[0] == ")":
                    return (self, signature[1:])
        elif token == "{":
            self = SignatureType("{")
            signature = signature[1:]
            (key_child, signature) = SignatureType._parse_next(signature)
            if not key_child or len(key_child.children):
                raise InvalidSignatureError("expected a simple type for dict entry key")
            self.children.append(key_child)
            (value_child, signature) = SignatureType._parse_next(signature)
            if not value_child:
                raise InvalidSignatureError("expected a value for dict entry")
            if not signature or signature[0] != "}":
                raise InvalidSignatureError('missing closing "}" for dict entry')
            self.children.append(value_child)
            return (self, signature[1:])

        # basic type
        return (SignatureType(token), signature[1:])

    def _verify_byte(self, body: Any) -> None:
        BYTE_MIN = 0x00
        BYTE_MAX = 0xFF
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus BYTE type "y" must be Python type "int", got {type(body)}'
            )
        if body < BYTE_MIN or body > BYTE_MAX:
            raise SignatureBodyMismatchError(
                f"DBus BYTE type must be between {BYTE_MIN} and {BYTE_MAX}"
            )

    def _verify_boolean(self, body: Any) -> None:
        if not isinstance(body, bool):
            raise SignatureBodyMismatchError(
                f'DBus BOOLEAN type "b" must be Python type "bool", got {type(body)}'
            )

    def _verify_int16(self, body: Any) -> None:
        INT16_MIN = -0x7FFF - 1
        INT16_MAX = 0x7FFF
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus INT16 type "n" must be Python type "int", got {type(body)}'
            )
        elif body > INT16_MAX or body < INT16_MIN:
            raise SignatureBodyMismatchError(
                f'DBus INT16 type "n" must be between {INT16_MIN} and {INT16_MAX}'
            )

    def _verify_uint16(self, body: Any) -> None:
        UINT16_MIN = 0
        UINT16_MAX = 0xFFFF
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus UINT16 type "q" must be Python type "int", got {type(body)}'
            )
        elif body > UINT16_MAX or body < UINT16_MIN:
            raise SignatureBodyMismatchError(
                f'DBus UINT16 type "q" must be between {UINT16_MIN} and {UINT16_MAX}'
            )

    def _verify_int32(self, body: int) -> None:
        INT32_MIN = -0x7FFFFFFF - 1
        INT32_MAX = 0x7FFFFFFF
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus INT32 type "i" must be Python type "int", got {type(body)}'
            )
        elif body > INT32_MAX or body < INT32_MIN:
            raise SignatureBodyMismatchError(
                f'DBus INT32 type "i" must be between {INT32_MIN} and {INT32_MAX}'
            )

    def _verify_uint32(self, body: Any) -> None:
        UINT32_MIN = 0
        UINT32_MAX = 0xFFFFFFFF
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus UINT32 type "u" must be Python type "int", got {type(body)}'
            )
        elif body > UINT32_MAX or body < UINT32_MIN:
            raise SignatureBodyMismatchError(
                f'DBus UINT32 type "u" must be between {UINT32_MIN} and {UINT32_MAX}'
            )

    def _verify_int64(self, body: Any) -> None:
        INT64_MAX = 9223372036854775807
        INT64_MIN = -INT64_MAX - 1
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus INT64 type "x" must be Python type "int", got {type(body)}'
            )
        elif body > INT64_MAX or body < INT64_MIN:
            raise SignatureBodyMismatchError(
                f'DBus INT64 type "x" must be between {INT64_MIN} and {INT64_MAX}'
            )

    def _verify_uint64(self, body: Any) -> None:
        UINT64_MIN = 0
        UINT64_MAX = 18446744073709551615
        if not isinstance(body, int):
            raise SignatureBodyMismatchError(
                f'DBus UINT64 type "t" must be Python type "int", got {type(body)}'
            )
        elif body > UINT64_MAX or body < UINT64_MIN:
            raise SignatureBodyMismatchError(
                f'DBus UINT64 type "t" must be between {UINT64_MIN} and {UINT64_MAX}'
            )

    def _verify_double(self, body: Any) -> None:
        if not isinstance(body, (float, int)):
            raise SignatureBodyMismatchError(
                f'DBus DOUBLE type "d" must be Python type "float" or "int", got {type(body)}'
            )

    def _verify_unix_fd(self, body: Any) -> None:
        try:
            self._verify_uint32(body)
        except SignatureBodyMismatchError:
            raise SignatureBodyMismatchError(
                'DBus UNIX_FD type "h" must be a valid UINT32'
            )

    def _verify_object_path(self, body: Any) -> None:
        if not is_object_path_valid(body):
            raise SignatureBodyMismatchError(
                'DBus OBJECT_PATH type "o" must be a valid object path'
            )

    def _verify_string(self, body: Any) -> None:
        if not isinstance(body, str):
            raise SignatureBodyMismatchError(
                f'DBus STRING type "s" must be Python type "str", got {type(body)}'
            )

    def _verify_signature(self, body: Any) -> None:
        # I guess we could run it through the SignatureTree parser instead
        if not isinstance(body, str):
            raise SignatureBodyMismatchError(
                f'DBus SIGNATURE type "g" must be Python type "str", got {type(body)}'
            )
        if len(body.encode()) > 0xFF:
            raise SignatureBodyMismatchError(
                'DBus SIGNATURE type "g" must be less than 256 bytes'
            )

    def _verify_array(self, body: Any) -> None:
        child_type = self.children[0]

        if child_type.token == "{":
            if not isinstance(body, dict):
                raise SignatureBodyMismatchError(
                    f'DBus ARRAY type "a" with DICT_ENTRY child must be Python type "dict", got {type(body)}'
                )
            for key, value in body.items():
                child_type.children[0].verify(key)
                child_type.children[1].verify(value)
        elif child_type.token == "y":
            if not isinstance(body, (bytearray, bytes)):
                raise SignatureBodyMismatchError(
                    f'DBus ARRAY type "a" with BYTE child must be Python type "bytes", got {type(body)}'
                )
                # no need to verify children
        else:
            if not isinstance(body, list):
                raise SignatureBodyMismatchError(
                    f'DBus ARRAY type "a" must be Python type "list", got {type(body)}'
                )
            for member in body:
                child_type.verify(member)

    def _verify_struct(self, body: Any) -> None:
        if not isinstance(body, (list, tuple)):
            raise SignatureBodyMismatchError(
                f'DBus STRUCT type "(" must be Python type "list" or "tuple", got {type(body)}'
            )

        if len(body) != len(self.children):
            raise SignatureBodyMismatchError(
                'DBus STRUCT type "(" must have Python list members equal to the number of struct type members'
            )

        for i, member in enumerate(body):
            self.children[i].verify(member)

    def _verify_variant(self, body: Any) -> None:
        # a variant signature and value is valid by construction
        if not isinstance(body, Variant):
            raise SignatureBodyMismatchError(
                f'DBus VARIANT type "v" must be Python type "Variant", got {type(body)}'
            )

    def verify(self, body: Any) -> bool:
        """Verify that the body matches this type.

        :returns: True if the body matches this type.
        :raises:
            :class:`SignatureBodyMismatchError` if the body does not match this type.
        """
        if body is None:
            raise SignatureBodyMismatchError('Cannot serialize Python type "None"')
        validator = self.validators.get(self.token)
        if validator:
            validator(self, body)
        else:
            raise Exception(f"cannot verify type with token {self.token}")

        return True

    validators: Dict[str, Callable[["SignatureType", Any], None]] = {
        "y": _verify_byte,
        "b": _verify_boolean,
        "n": _verify_int16,
        "q": _verify_uint16,
        "i": _verify_int32,
        "u": _verify_uint32,
        "x": _verify_int64,
        "t": _verify_uint64,
        "d": _verify_double,
        "h": _verify_uint32,
        "o": _verify_string,
        "s": _verify_string,
        "g": _verify_signature,
        "a": _verify_array,
        "(": _verify_struct,
        "v": _verify_variant,
    }


class SignatureTree:
    """A class that represents a signature as a tree structure for conveniently
    working with DBus signatures.

    This class will not normally be used directly by the user.

    :ivar types: A list of parsed complete types.
    :vartype types: list(:class:`SignatureType`)

    :ivar ~.signature: The signature of this signature tree.
    :vartype ~.signature: str

    :raises:
        :class:`InvalidSignatureError` if the given signature is not valid.
    """

    __slots__ = ("signature", "types")

    def __init__(self, signature: str = "") -> None:
        self.signature = signature

        self.types: List[SignatureType] = []

        if len(signature) > 0xFF:
            raise InvalidSignatureError("A signature must be less than 256 characters")

        while signature:
            (type_, signature) = SignatureType._parse_next(signature)
            self.types.append(type_)

    def __eq__(self, other: Any) -> bool:
        if type(other) is SignatureTree:
            return self.signature == other.signature
        return super().__eq__(other)

    def verify(self, body: List[Any]) -> bool:
        """Verifies that the give body matches this signature tree

        :param body: the body to verify for this tree
        :type body: list(Any)

        :returns: True if the signature matches the body or an exception if not.

        :raises:
            :class:`SignatureBodyMismatchError` if the signature does not match the body.
        """
        if not isinstance(body, list):
            raise SignatureBodyMismatchError(
                f"The body must be a list (got {type(body)})"
            )
        if len(body) != len(self.types):
            raise SignatureBodyMismatchError(
                f"The body has the wrong number of types (got {len(body)}, expected {len(self.types)})"
            )
        for i, type_ in enumerate(self.types):
            type_.verify(body[i])

        return True


class Variant:
    """A class to represent a DBus variant (type "v").

    This class is used in message bodies to represent variants. The user can
    expect a value in the body with type "v" to use this class and can
    construct this class directly for use in message bodies sent over the bus.

    :ivar signature: The signature for this variant. Must be a single complete type.
    :vartype signature: str or SignatureTree or SignatureType

    :ivar value: The value of this variant. Must correspond to the signature.
    :vartype value: Any

    :raises:
        :class:`InvalidSignatureError` if the signature is not valid.
        :class:`SignatureBodyMismatchError` if the signature does not match the body.
    """

    __slots__ = ("type", "signature", "value")

    def __init__(
        self,
        signature: Union[str, SignatureTree, SignatureType],
        value: Any,
        verify: bool = True,
    ) -> None:
        """Init a new Variant."""
        self._init_variant(signature, value, verify)

    def _init_variant(
        self,
        signature: Union[str, SignatureTree, SignatureType],
        value: Any,
        verify: bool,
    ) -> None:
        if type(signature) is SignatureTree:
            signature_tree = signature
            self.signature = signature_tree.signature
            self.type = signature_tree.types[0]
        elif type(signature) is SignatureType:
            signature_tree = None
            self.signature = signature.signature
            self.type = signature
        elif type(signature) is str:
            signature_tree = get_signature_tree(signature)
            self.signature = signature
            self.type = signature_tree.types[0]
        else:
            raise TypeError(
                "signature must be a SignatureTree, SignatureType, or a string"
            )
        self.value = value
        if verify:
            if signature_tree and len(signature_tree.types) != 1:
                raise ValueError(
                    "variants must have a signature for a single complete type"
                )
            self.type.verify(value)

    def __eq__(self, other: Any) -> bool:
        if type(other) is Variant:
            return self.signature == other.signature and self.value == other.value
        return super().__eq__(other)

    def __repr__(self) -> str:
        return "<dbus_fast.signature.Variant ('{}', {})>".format(
            self.type.signature, self.value
        )


get_signature_tree = lru_cache(maxsize=None)(SignatureTree)
"""Get a signature tree for the given signature.

:param signature: The signature to get a tree for.
:type signature: str

:returns: The signature tree for the given signature.
:rtype: :class:`SignatureTree`
"""



import pytest

from dbus_fast import SignatureBodyMismatchError, SignatureTree, Variant
from dbus_fast._private.util import signature_contains_type


def assert_simple_type(signature, type_):
    assert type_.token == signature
    assert type_.signature == signature
    assert len(type_.children) == 0


def test_simple():
    tree = SignatureTree("s")
    assert len(tree.types) == 1
    assert_simple_type("s", tree.types[0])


def test_multiple_simple():
    tree = SignatureTree("sss")
    assert len(tree.types) == 3
    for i in range(0, 3):
        assert_simple_type("s", tree.types[i])


def test_array():
    tree = SignatureTree("as")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "as"
    assert child.token == "a"
    assert len(child.children) == 1
    assert_simple_type("s", child.children[0])


def test_array_multiple():
    tree = SignatureTree("asasass")
    assert len(tree.types) == 4
    assert_simple_type("s", tree.types[3])
    for i in range(0, 3):
        array_child = tree.types[i]
        assert array_child.token == "a"
        assert array_child.signature == "as"
        assert len(array_child.children) == 1
        assert_simple_type("s", array_child.children[0])


def test_array_nested():
    tree = SignatureTree("aas")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.token == "a"
    assert child.signature == "aas"
    assert len(child.children) == 1
    nested_child = child.children[0]
    assert nested_child.token == "a"
    assert nested_child.signature == "as"
    assert len(nested_child.children) == 1
    assert_simple_type("s", nested_child.children[0])


def test_simple_struct():
    tree = SignatureTree("(sss)")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "(sss)"
    assert len(child.children) == 3
    for i in range(0, 3):
        assert_simple_type("s", child.children[i])


def test_nested_struct():
    tree = SignatureTree("(s(s(s)))")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "(s(s(s)))"
    assert child.token == "("
    assert len(child.children) == 2
    assert_simple_type("s", child.children[0])
    first_nested = child.children[1]
    assert first_nested.token == "("
    assert first_nested.signature == "(s(s))"
    assert len(first_nested.children) == 2
    assert_simple_type("s", first_nested.children[0])
    second_nested = first_nested.children[1]
    assert second_nested.token == "("
    assert second_nested.signature == "(s)"
    assert len(second_nested.children) == 1
    assert_simple_type("s", second_nested.children[0])


def test_struct_multiple():
    tree = SignatureTree("(s)(s)(s)")
    assert len(tree.types) == 3
    for i in range(0, 3):
        child = tree.types[0]
        assert child.token == "("
        assert child.signature == "(s)"
        assert len(child.children) == 1
        assert_simple_type("s", child.children[0])


def test_array_of_structs():
    tree = SignatureTree("a(ss)")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.token == "a"
    assert child.signature == "a(ss)"
    assert len(child.children) == 1
    struct_child = child.children[0]
    assert struct_child.token == "("
    assert struct_child.signature == "(ss)"
    assert len(struct_child.children) == 2
    for i in range(0, 2):
        assert_simple_type("s", struct_child.children[i])


def test_dict_simple():
    tree = SignatureTree("a{ss}")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "a{ss}"
    assert child.token == "a"
    assert len(child.children) == 1
    dict_child = child.children[0]
    assert dict_child.token == "{"
    assert dict_child.signature == "{ss}"
    assert len(dict_child.children) == 2
    assert_simple_type("s", dict_child.children[0])
    assert_simple_type("s", dict_child.children[1])


def test_dict_of_structs():
    tree = SignatureTree("a{s(ss)}")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.token == "a"
    assert child.signature == "a{s(ss)}"
    assert len(child.children) == 1
    dict_child = child.children[0]
    assert dict_child.token == "{"
    assert dict_child.signature == "{s(ss)}"
    assert len(dict_child.children) == 2
    assert_simple_type("s", dict_child.children[0])
    struct_child = dict_child.children[1]
    assert struct_child.token == "("
    assert struct_child.signature == "(ss)"
    assert len(struct_child.children) == 2
    for i in range(0, 2):
        assert_simple_type("s", struct_child.children[i])


def test_contains_type():
    tree = SignatureTree("h")
    assert signature_contains_type(tree, [0], "h")
    assert not signature_contains_type(tree, [0], "u")

    tree = SignatureTree("ah")
    assert signature_contains_type(tree, [[0]], "h")
    assert signature_contains_type(tree, [[0]], "a")
    assert not signature_contains_type(tree, [[0]], "u")

    tree = SignatureTree("av")
    body = [
        [
            Variant("u", 0),
            Variant("i", 0),
            Variant("x", 0),
            Variant("v", Variant("s", "hi")),
        ]
    ]
    assert signature_contains_type(tree, body, "u")
    assert signature_contains_type(tree, body, "x")
    assert signature_contains_type(tree, body, "v")
    assert signature_contains_type(tree, body, "s")
    assert not signature_contains_type(tree, body, "o")

    tree = SignatureTree("a{sv}")
    body = {
        "foo": Variant("h", 0),
        "bar": Variant("i", 0),
        "bat": Variant("x", 0),
        "baz": Variant("v", Variant("o", "/hi")),
    }
    for expected in "hixvso":
        assert signature_contains_type(tree, [body], expected)
    assert not signature_contains_type(tree, [body], "b")


def test_invalid_variants():
    tree = SignatureTree("a{sa{sv}}")
    s_con = {
        "type": "802-11-wireless",
        "uuid": "1234",
        "id": "SSID",
    }

    s_wifi = {
        "ssid": "SSID",
        "mode": "infrastructure",
        "hidden": True,
    }

    s_wsec = {
        "key-mgmt": "wpa-psk",
        "auth-alg": "open",
        "psk": "PASSWORD",
    }

    s_ip4 = {"method": "auto"}
    s_ip6 = {"method": "auto"}

    con = {
        "connection": s_con,
        "802-11-wireless": s_wifi,
        "802-11-wireless-security": s_wsec,
        "ipv4": s_ip4,
        "ipv6": s_ip6,
    }

    with pytest.raises(SignatureBodyMismatchError):
        tree.verify([con])


def test_variant_signature_type():
    tree = SignatureTree("as")
    var = Variant(tree.types[0], ["foo", "bar"])
    assert var.type == tree.types[0]
    assert var.value == ["foo", "bar"]
    assert var.signature == "as"

    with pytest.raises(SignatureBodyMismatchError):
        Variant(tree.types[0], "wrong")


def test_struct_accepts_tuples_or_lists():
    tree = SignatureTree("(s)")
    tree.verify([("ok",)])
    tree.verify([["ok"]])



from typing import Optional, Union


class SignatureBodyMismatchError(ValueError):
    pass


class InvalidSignatureError(ValueError):
    pass


class InvalidAddressError(ValueError):
    pass


class AuthError(Exception):
    pass


class InvalidMessageError(ValueError):
    pass


class InvalidIntrospectionError(ValueError):
    pass


class InterfaceNotFoundError(Exception):
    pass


class SignalDisabledError(Exception):
    pass


class InvalidBusNameError(TypeError):
    def __init__(self, name: str) -> None:
        super().__init__(f"invalid bus name: {name}")


class InvalidObjectPathError(TypeError):
    def __init__(self, path: str) -> None:
        super().__init__(f"invalid object path: {path}")


class InvalidInterfaceNameError(TypeError):
    def __init__(self, name: str) -> None:
        super().__init__(f"invalid interface name: {name}")


class InvalidMemberNameError(TypeError):
    def __init__(self, member: str) -> None:
        super().__init__(f"invalid member name: {member}")


from .constants import ErrorType, MessageType
from .message import Message
from .validators import assert_interface_name_valid


class DBusError(Exception):
    def __init__(
        self, type_: Union[ErrorType, str], text: str, reply: Optional[Message] = None
    ) -> None:
        super().__init__(text)

        if type(type_) is ErrorType:
            type_ = type_.value

        assert_interface_name_valid(type_)  # type: ignore[arg-type]
        if reply is not None and type(reply) is not Message:
            raise TypeError("reply must be of type Message")

        self.type = type_
        self.text = text
        self.reply = reply

    @staticmethod
    def _from_message(msg: Message) -> "DBusError":
        assert msg.message_type == MessageType.ERROR
        return DBusError(msg.error_name or "unknown", msg.body[0], reply=msg)

    def _as_message(self, msg: Message) -> Message:
        return Message.new_error(msg, self.type, self.text)



import pytest

from dbus_fast import aio
from dbus_fast.service import ServiceInterface


class ExampleInterface(ServiceInterface):
    def __init__(self):
        super().__init__("test.interface")


@pytest.mark.asyncio
async def test_fast_disconnect():
    bus_name = "aio.client.test.methods"

    bus = await aio.MessageBus().connect()
    bus2 = await aio.MessageBus().connect()
    await bus.request_name(bus_name)
    service_interface = ExampleInterface()
    bus.export("/test/path", service_interface)
    introspection = await bus2.introspect(bus_name, "/test/path")
    bus2.get_proxy_object(bus_name, "/test/path", introspection)
    bus2.disconnect()
    bus.disconnect()
    await bus.wait_for_disconnect()
    await bus2.wait_for_disconnect()



import xml.etree.ElementTree as ET
from typing import List, Union

from .. import introspection as intr
from ..constants import ErrorType
from ..errors import DBusError
from ..message import Message
from ..message_bus import BaseMessageBus
from ..proxy_object import BaseProxyInterface, BaseProxyObject
from ..signature import Variant
from ..unpack import unpack_variants as unpack

# glib is optional
try:
    from gi.repository import GLib
except ImportError:
    pass


class ProxyInterface(BaseProxyInterface):
    """A class representing a proxy to an interface exported on the bus by
    another client for the GLib :class:`MessageBus <dbus_fast.glib.MessageBus>`
    implementation.

    This class is not meant to be constructed directly by the user. Use
    :func:`ProxyObject.get_interface()
    <dbus_fast.glib.ProxyObject.get_interface>` on a GLib proxy
    object to get a proxy interface.

    This class exposes methods to call DBus methods, listen to signals, and get
    and set properties on the interface that are created dynamically based on
    the introspection data passed to the proxy object that made this proxy
    interface.

    A *method call* takes this form:

    .. code-block:: python3

        def callback(error: Exception, result: list(Any)):
            pass

        interface.call_[METHOD](*args, callback)
        result = interface.call_[METHOD]_sync(*args)

    Where ``METHOD`` is the name of the method converted to snake case.

    To call a method, provide ``*args`` that correspond to the *in args* of the
    introspection method definition.

    To *asynchronously* call a method, provide a callback that takes an error
    as the first argument and a list as the second argument. If the call
    completed successfully, ``error`` will be :class:`None`. If the service
    returns an error, it will be a :class:`DBusError <dbus_fast.DBusError>`
    with information about the error returned from the bus. The result will be
    a list of values that correspond to the *out args* of the introspection
    method definition.

    To *synchronously* call a method, use the ``call_[METHOD]_sync()`` form.
    The ``result`` corresponds to the *out arg* of the introspection method
    definition. If the method has more than one otu arg, they are returned
    within a :class:`list`.

    To *listen to a signal* use this form:

    .. code-block:: python3

        interface.on_[SIGNAL](callback)

    To *stop listening to a signal* use this form:

    .. code-block:: python3

        interface.off_[SIGNAL](callback)

    Where ``SIGNAL`` is the name of the signal converted to snake case.

    DBus signals are exposed with an event-callback interface. The provided
    ``callback`` will be called when the signal is emitted with arguments that
    correspond to the *out args* of the interface signal definition.

    To *get or set a property* use this form:

    .. code-block:: python3

        def get_callback(error: Exception, value: Any):
            pass

        def set_callback(error: Exception)
            pass

        interface.get_[PROPERTY](get_callback)
        value: Any = interface.get_[PROPERTY]_sync()

        interface.set_[PROPERTY](set_callback)
        interface.set_[PROPERTY]_sync(value)

    Where ``PROPERTY`` is the name of the property converted to snake case.

    The ``value`` must correspond to the type of the property in the interface
    definition.

    To asynchronously get or set a property, provide a callback that takes an
    :class:`Exception` as the first argument. If the call completed
    successfully, ``error`` will be :class:`None`. If the service returns an
    error, it will be a :class:`DBusError <dbus_fast.DBusError>` with
    information about the error returned from the bus.

    If the service returns an error for a synchronous DBus call, a
    :class:`DBusError <dbus_fast.DBusError>` will be raised with information
    about the error.
    """

    def _add_method(self, intr_method):
        in_len = len(intr_method.in_args)
        out_len = len(intr_method.out_args)

        def method_fn(*args, unpack_variants: bool = False):
            if len(args) != in_len + 1:
                raise TypeError(
                    f"method {intr_method.name} expects {in_len} arguments and a callback (got {len(args)} args)"
                )

            args = list(args)
            # TODO type check: this callback takes two parameters
            # (MessageBus.check_callback(cb))
            callback = args.pop()

            def call_notify(msg, err):
                if err:
                    callback([], err)
                    return

                try:
                    BaseProxyInterface._check_method_return(
                        msg, intr_method.out_signature
                    )
                except DBusError as e:
                    err = e

                if unpack_variants:
                    callback(unpack(msg.body), err)
                else:
                    callback(msg.body, err)

            self.bus.call(
                Message(
                    destination=self.bus_name,
                    path=self.path,
                    interface=self.introspection.name,
                    member=intr_method.name,
                    signature=intr_method.in_signature,
                    body=list(args),
                ),
                call_notify,
            )

        def method_fn_sync(*args, unpack_variants: bool = False):
            main = GLib.MainLoop()
            call_error = None
            call_body = None

            def callback(body, err):
                nonlocal call_error
                nonlocal call_body
                call_error = err
                call_body = body
                main.quit()

            method_fn(*args, callback)

            main.run()

            if call_error:
                raise call_error

            if not out_len:
                return None

            if unpack_variants:
                call_body = unpack(call_body)

            if out_len == 1:
                return call_body[0]
            return call_body

        method_name = f"call_{BaseProxyInterface._to_snake_case(intr_method.name)}"
        method_name_sync = f"{method_name}_sync"

        setattr(self, method_name, method_fn)
        setattr(self, method_name_sync, method_fn_sync)

    def _add_property(self, intr_property):
        def property_getter(callback, *, unpack_variants: bool = False):
            def call_notify(msg, err):
                if err:
                    callback(None, err)
                    return

                try:
                    BaseProxyInterface._check_method_return(msg)
                except Exception as e:
                    callback(None, e)
                    return

                variant = msg.body[0]
                if variant.signature != intr_property.signature:
                    err = DBusError(
                        ErrorType.CLIENT_ERROR,
                        'property returned unexpected signature "{variant.signature}"',
                        msg,
                    )
                    callback(None, err)
                    return
                if unpack_variants:
                    callback(unpack(variant.value), None)
                else:
                    callback(variant.value, None)

            self.bus.call(
                Message(
                    destination=self.bus_name,
                    path=self.path,
                    interface="org.freedesktop.DBus.Properties",
                    member="Get",
                    signature="ss",
                    body=[self.introspection.name, intr_property.name],
                ),
                call_notify,
            )

        def property_getter_sync(*, unpack_variants: bool = False):
            property_value = None
            reply_error = None

            main = GLib.MainLoop()

            def callback(value, err):
                nonlocal property_value
                nonlocal reply_error
                property_value = value
                reply_error = err
                main.quit()

            property_getter(callback)
            main.run()
            if reply_error:
                raise reply_error
            if unpack_variants:
                return unpack(property_value)
            return property_value

        def property_setter(value, callback):
            def call_notify(msg, err):
                if err:
                    callback(None, err)
                    return
                try:
                    BaseProxyInterface._check_method_return(msg)
                except Exception as e:
                    callback(None, e)
                    return

                return callback(None, None)

            variant = Variant(intr_property.signature, value)
            self.bus.call(
                Message(
                    destination=self.bus_name,
                    path=self.path,
                    interface="org.freedesktop.DBus.Properties",
                    member="Set",
                    signature="ssv",
                    body=[self.introspection.name, intr_property.name, variant],
                ),
                call_notify,
            )

        def property_setter_sync(val):
            reply_error = None

            main = GLib.MainLoop()

            def callback(value, err):
                nonlocal reply_error
                reply_error = err
                main.quit()

            property_setter(val, callback)
            main.run()
            if reply_error:
                raise reply_error
            return None

        snake_case = super()._to_snake_case(intr_property.name)
        setattr(self, f"get_{snake_case}", property_getter)
        setattr(self, f"get_{snake_case}_sync", property_getter_sync)
        setattr(self, f"set_{snake_case}", property_setter)
        setattr(self, f"set_{snake_case}_sync", property_setter_sync)


class ProxyObject(BaseProxyObject):
    """The proxy object implementation for the asyncio :class:`MessageBus <dbus_fast.aio.MessageBus>`.

    For more information, see the :class:`BaseProxyObject <dbus_fast.proxy_object.BaseProxyObject>`.
    """

    def __init__(
        self,
        bus_name: str,
        path: str,
        introspection: Union[intr.Node, str, ET.Element],
        bus: BaseMessageBus,
    ):
        super().__init__(bus_name, path, introspection, bus, ProxyInterface)

    def get_interface(self, name: str) -> ProxyInterface:
        return super().get_interface(name)

    def get_children(self) -> List["ProxyObject"]:
        return super().get_children()



import io
import timeit

from dbus_fast._private.unmarshaller import Unmarshaller

bluez_passive_message = (
    b"l\1\0\1(\0\0\0F\1\0\0\255\0\0\0\1\1o\0:\0\0\0/org/bluez/hci0/dev_BD_24_6F_"
    b"85_AA_61/service001a/char001b\0\0\0\0\0\0\2\1s\0\35\0\0\0org.bluez.GattCha"
    b"racteristic1\0\0\0\3\1s\0\n\0\0\0WriteValue\0\0\0\0\0\0\6\1s\0\t\0\0\0org."
    b"bluez\0\0\0\0\0\0\0\10\1g\0\7aya{sv}\0\0\0\0\3\0\0\0\357\1w\0\30\0\0\0\0\0"
    b"\0\0\4\0\0\0type\0\1s\0\7\0\0\0command\0"
)


stream = io.BytesIO(bluez_passive_message)

unmarshaller = Unmarshaller(stream)


def unmarhsall_bluez_rssi_message():
    stream.seek(0)
    unmarshaller.unmarshall()


count = 3000000
time = timeit.Timer(unmarhsall_bluez_rssi_message).timeit(count)
print(f"Unmarshalling {count} bluetooth write value messages took {time} seconds")



#!/usr/bin/env python3

# In order for this to work a local tcp connection to the DBus a port
# must be opened to forward to the dbus socket file. The easiest way
# to achieve this is using "socat":
# socat TCP-LISTEN:55556,reuseaddr,fork,range=127.0.0.1/32 UNIX-CONNECT:$(echo $DBUS_SESSION_BUS_ADDRESS | sed 's/unix:path=//g')
# For actual DBus transport over network the authentication might
# be a further problem. More information here:
# https://dbus.freedesktop.org/doc/dbus-specification.html#auth-mechanisms

import os
import sys

sys.path.append(os.path.abspath(os.path.dirname(__file__) + "/.."))

import asyncio

from dbus_fast.aio import MessageBus


async def main():
    bus = await MessageBus(bus_address="tcp:host=127.0.0.1,port=55556").connect()
    introspection = await bus.introspect(
        "org.freedesktop.Notifications", "/org/freedesktop/Notifications"
    )
    obj = bus.get_proxy_object(
        "org.freedesktop.Notifications", "/org/freedesktop/Notifications", introspection
    )
    notification = obj.get_interface("org.freedesktop.Notifications")
    await notification.call_notify(
        "test.py", 0, "", "DBus Test", "Test notification", [""], dict(), 5000
    )


asyncio.run(main())
